import asyncio

from mirascope import llm


async def recommend_book(genre: str) -> llm.AsyncResponse:
    model: llm.Model = llm.use_model(provider="openai", model_id="gpt-5")
    message = llm.messages.user(f"Please recommend a book in {genre}.")
    return await model.call_async(messages=[message])


async def main():
    response: llm.AsyncResponse = await recommend_book("fantasy")
    print(response.pretty())


if __name__ == "__main__":
    asyncio.run(main())



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    FinishReason,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:completions",
            "model_id": "gpt-4o",
            "params": {"max_tokens": 50},
            "finish_reason": FinishReason.MAX_TOKENS,
            "messages": [
                UserMessage(content=[Text(text="List all U.S. states.")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
Certainly! Here is a list of all 50 U.S. states:

1. Alabama
2. Alaska
3. Arizona
4. Arkansas
5. California
6. Colorado
7. Connecticut
8. Delaware
9. Florida\
"""
                        )
                    ],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message={
                        "content": """\
Certainly! Here is a list of all 50 U.S. states:

1. Alabama
2. Alaska
3. Arizona
4. Arkansas
5. California
6. Colorado
7. Connecticut
8. Delaware
9. Florida\
""",
                        "role": "assistant",
                        "annotations": [],
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:completions",
            "model_id": "gpt-4o",
            "params": {"max_tokens": 50},
            "finish_reason": FinishReason.MAX_TOKENS,
            "messages": [
                UserMessage(content=[Text(text="List all U.S. states.")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
Sure! Here is a list of all 50 U.S. states:

1. Alabama
2. Alaska
3. Arizona
4. Arkansas
5. California
6. Colorado
7. Connecticut
8. Delaware
9. Florida\
"""
                        )
                    ],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message={
                        "content": """\
Sure! Here is a list of all 50 U.S. states:

1. Alabama
2. Alaska
3. Arizona
4. Arkansas
5. California
6. Colorado
7. Connecticut
8. Delaware
9. Florida\
""",
                        "role": "assistant",
                        "annotations": [],
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:completions",
            "model_id": "gpt-4o",
            "finish_reason": FinishReason.MAX_TOKENS,
            "messages": [
                UserMessage(content=[Text(text="List all U.S. states.")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
Certainly! Here is a list of all 50 U.S. states:

1. Alabama
2. Alaska
3. Arizona
4. Arkansas
5. California
6. Colorado
7. Connecticut
8. Delaware
9. Florida\
"""
                        )
                    ],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [],
            "n_chunks": 52,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:completions",
            "model_id": "gpt-4o",
            "finish_reason": FinishReason.MAX_TOKENS,
            "messages": [
                UserMessage(content=[Text(text="List all U.S. states.")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
Sure! Here is a list of all 50 U.S. states:

1. Alabama
2. Alaska
3. Arizona
4. Arkansas
5. California
6. Colorado
7. Connecticut
8. Delaware
9. Florida\
"""
                        )
                    ],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [],
            "n_chunks": 52,
        }
    }
)



"""Mirascope exception hierarchy for unified error handling across providers."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .clients import ModelId, Provider
    from .formatting import FormattingMode


class MirascopeError(Exception):
    """Base exception for all Mirascope errors."""

    original_exception: Exception | None


class APIError(MirascopeError):
    """Base class for API-related errors."""

    status_code: int | None


class ConnectionError(MirascopeError):
    """Raised when unable to connect to the API (network issues, timeouts)."""


class AuthenticationError(APIError):
    """Raised for authentication failures (401, invalid API keys)."""


class PermissionError(APIError):
    """Raised for permission/authorization failures (403)."""


class BadRequestError(APIError):
    """Raised for malformed requests (400, 422)."""


class NotFoundError(APIError):
    """Raised when requested resource is not found (404)."""


class ToolNotFoundError(MirascopeError):
    """Raised if a tool_call cannot be converted to any corresponding tool."""


class FeatureNotSupportedError(MirascopeError):
    """Raised if a Mirascope feature is unsupported by chosen provider.

    If compatibility is model-specific, then `model_id` should be specified.
    If the feature is not supported by the provider at all, then it may be `None`."""

    provider: "Provider"
    model_id: "ModelId | None"
    feature: str

    def __init__(
        self,
        feature: str,
        provider: "Provider",
        model_id: "ModelId | None" = None,
        message: str | None = None,
    ) -> None:
        if message is None:
            model_msg = f" for model '{model_id}'" if model_id is not None else ""
            message = f"Feature '{feature}' is not supported by provider '{provider}'{model_msg}"
        super().__init__(message)
        self.feature = feature
        self.provider = provider
        self.model_id = model_id


class FormattingModeNotSupportedError(FeatureNotSupportedError):
    """Raised when trying to use a formatting mode that is not supported by the chosen model."""

    formatting_mode: "FormattingMode"

    def __init__(
        self,
        formatting_mode: "FormattingMode",
        provider: "Provider",
        model_id: "ModelId | None" = None,
        message: str | None = None,
    ) -> None:
        if message is None:
            model_msg = f" for model '{model_id}'" if model_id is not None else ""
            message = f"Formatting mode '{formatting_mode}' is not supported by provider '{provider}'{model_msg}"
        super().__init__(
            feature=f"formatting_mode:{formatting_mode}",
            provider=provider,
            model_id=model_id,
            message=message,
        )
        self.formatting_mode = formatting_mode


class RateLimitError(APIError):
    """Raised when rate limits are exceeded (429)."""


class ServerError(APIError):
    """Raised for server-side errors (500+)."""


class TimeoutError(MirascopeError):
    """Raised when requests timeout or deadline exceeded."""



"""End-to-end tests for LLM tool usage, including parallel tool usage and continuation."""

import pytest

from mirascope import llm
from tests.e2e.conftest import PROVIDER_MODEL_ID_PAIRS
from tests.utils import (
    Snapshot,
    snapshot_test,
)

PASSWORD_MAP = {"mellon": "Welcome to Moria!", "radiance": "Life before Death"}

# ============= SYNC TESTS =============


@pytest.mark.parametrize("provider, model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
def test_call_with_tools_sync(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test synchronous tool call without context."""

    @llm.tool
    def secret_retrieval_tool(password: str) -> str:
        """A tool that requires a password to retrieve a secret."""
        return PASSWORD_MAP.get(password, "Invalid password!")

    @llm.call(
        provider=provider,
        model_id=model_id,
        tools=[secret_retrieval_tool],
    )
    def call(passwords: list[str]) -> list[llm.Message]:
        return [
            llm.messages.system("Use parallel tool calling."),
            llm.messages.user(
                f"Please retrieve the secrets associated with each of these passwords: {','.join(passwords)}"
            ),
        ]

    with snapshot_test(snapshot) as snap:
        response = call(["mellon", "radiance"])
        assert len(response.tool_calls) == 2, (
            f"Expected response to have two tool calls: {response.pretty()}"
        )

        tool_outputs = response.execute_tools()
        response = response.resume(tool_outputs)

        snap.set_response(response)
        pretty = response.pretty()
        assert "Moria" in pretty, f"Expected 'Moria' to be in response: {pretty}"
        assert "Life before Death" in pretty, (
            f"Expected 'Life before Death' to be in response: {pretty}"
        )


@pytest.mark.parametrize("provider, model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
def test_call_with_tools_sync_context(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test synchronous tool call with context."""

    @llm.tool
    def secret_retrieval_tool(ctx: llm.Context[dict[str, str]], password: str) -> str:
        """A tool that requires a password to retrieve a secret."""
        return ctx.deps.get(password, "Invalid password!")

    @llm.call(
        provider=provider,
        model_id=model_id,
        tools=[secret_retrieval_tool],
    )
    def call(
        ctx: llm.Context[dict[str, str]], passwords: list[str]
    ) -> list[llm.Message]:
        return [
            llm.messages.system("Use parallel tool calling."),
            llm.messages.user(
                f"Please retrieve the secrets associated with each of these passwords: {','.join(passwords)}"
            ),
        ]

    with snapshot_test(snapshot) as snap:
        ctx = llm.Context(deps=PASSWORD_MAP)
        response = call(ctx, ["mellon", "radiance"])
        assert len(response.tool_calls) == 2, (
            f"Expected response to have two tool calls: {response.pretty()}"
        )

        tool_outputs = response.execute_tools(
            ctx,
        )
        response = response.resume(ctx, tool_outputs)

        snap.set_response(response)
        pretty = response.pretty()
        assert "Moria" in pretty, f"Expected 'Moria' to be in response: {pretty}"
        assert "Life before Death" in pretty, (
            f"Expected 'Life before Death' to be in response: {pretty}"
        )


# ============= ASYNC TESTS =============


@pytest.mark.parametrize("provider, model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_tools_async(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test asynchronous tool call without context."""

    @llm.tool
    async def secret_retrieval_tool(password: str) -> str:
        """A tool that requires a password to retrieve a secret."""
        return PASSWORD_MAP.get(password, "Invalid password!")

    @llm.call(
        provider=provider,
        model_id=model_id,
        tools=[secret_retrieval_tool],
    )
    async def call(passwords: list[str]) -> list[llm.Message]:
        return [
            llm.messages.system("Use parallel tool calling."),
            llm.messages.user(
                f"Please retrieve the secrets associated with each of these passwords: {','.join(passwords)}"
            ),
        ]

    with snapshot_test(snapshot) as snap:
        response = await call(["mellon", "radiance"])
        assert len(response.tool_calls) == 2, (
            f"Expected response to have two tool calls: {response.pretty()}"
        )

        tool_outputs = await response.execute_tools()
        response = await response.resume(tool_outputs)

        snap.set_response(response)
        pretty = response.pretty()
        assert "Moria" in pretty, f"Expected 'Moria' to be in response: {pretty}"
        assert "Life before Death" in pretty, (
            f"Expected 'Life before Death' to be in response: {pretty}"
        )


@pytest.mark.parametrize("provider, model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_tools_async_context(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test asynchronous tool call with context."""

    @llm.tool
    async def secret_retrieval_tool(
        ctx: llm.Context[dict[str, str]], password: str
    ) -> str:
        """A tool that requires a password to retrieve a secret."""
        return ctx.deps.get(password, "Invalid password!")

    @llm.call(
        provider=provider,
        model_id=model_id,
        tools=[secret_retrieval_tool],
    )
    async def call(
        ctx: llm.Context[dict[str, str]], passwords: list[str]
    ) -> list[llm.Message]:
        return [
            llm.messages.system("Use parallel tool calling."),
            llm.messages.user(
                f"Please retrieve the secrets associated with each of these passwords: {','.join(passwords)}"
            ),
        ]

    with snapshot_test(snapshot) as snap:
        ctx = llm.Context(deps=PASSWORD_MAP)
        response = await call(ctx, ["mellon", "radiance"])
        assert len(response.tool_calls) == 2, (
            f"Expected response to have two tool calls: {response.pretty()}"
        )

        tool_outputs = await response.execute_tools(ctx)
        response = await response.resume(ctx, tool_outputs)

        snap.set_response(response)
        pretty = response.pretty()
        assert "Moria" in pretty, f"Expected 'Moria' to be in response: {pretty}"
        assert "Life before Death" in pretty, (
            f"Expected 'Life before Death' to be in response: {pretty}"
        )


# ============= STREAM TESTS =============


@pytest.mark.parametrize("provider, model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
def test_call_with_tools_stream(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test streaming tool call without context."""

    @llm.tool
    def secret_retrieval_tool(password: str) -> str:
        """A tool that requires a password to retrieve a secret."""
        return PASSWORD_MAP.get(password, "Invalid password!")

    @llm.call(
        provider=provider,
        model_id=model_id,
        tools=[secret_retrieval_tool],
    )
    def call(passwords: list[str]) -> list[llm.Message]:
        return [
            llm.messages.system("Use parallel tool calling."),
            llm.messages.user(
                f"Please retrieve the secrets associated with each of these passwords: {','.join(passwords)}"
            ),
        ]

    with snapshot_test(snapshot) as snap:
        response = call.stream(["mellon", "radiance"])
        response.finish()

        assert len(response.tool_calls) == 2, (
            f"Expected response to have two tool calls: {response.pretty()}"
        )

        tool_outputs = response.execute_tools()
        response = response.resume(tool_outputs)
        response.finish()

        snap.set_response(response)
        pretty = response.pretty()
        assert "Moria" in pretty, f"Expected 'Moria' to be in response: {pretty}"
        assert "Life before Death" in pretty, (
            f"Expected 'Life before Death' to be in response: {pretty}"
        )


@pytest.mark.parametrize("provider, model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
def test_call_with_tools_stream_context(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test streaming tool call with context."""

    @llm.tool
    def secret_retrieval_tool(ctx: llm.Context[dict[str, str]], password: str) -> str:
        """A tool that requires a password to retrieve a secret."""
        return ctx.deps.get(password, "Invalid password!")

    @llm.call(
        provider=provider,
        model_id=model_id,
        tools=[secret_retrieval_tool],
    )
    def call(
        ctx: llm.Context[dict[str, str]], passwords: list[str]
    ) -> list[llm.Message]:
        return [
            llm.messages.system("Use parallel tool calling."),
            llm.messages.user(
                f"Please retrieve the secrets associated with each of these passwords: {','.join(passwords)}"
            ),
        ]

    with snapshot_test(snapshot) as snap:
        ctx = llm.Context(deps=PASSWORD_MAP)
        response = call.stream(ctx, ["mellon", "radiance"])
        response.finish()

        assert len(response.tool_calls) == 2, (
            f"Expected response to have two tool calls: {response.pretty()}"
        )

        tool_outputs = response.execute_tools(ctx)
        response = response.resume(ctx, tool_outputs)
        response.finish()

        snap.set_response(response)
        pretty = response.pretty()
        assert "Moria" in pretty, f"Expected 'Moria' to be in response: {pretty}"
        assert "Life before Death" in pretty, (
            f"Expected 'Life before Death' to be in response: {pretty}"
        )


# ============= ASYNC STREAM TESTS =============


@pytest.mark.parametrize("provider, model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_tools_async_stream(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test async streaming tool call without context."""

    @llm.tool
    async def secret_retrieval_tool(password: str) -> str:
        """A tool that requires a password to retrieve a secret."""
        return PASSWORD_MAP.get(password, "Invalid password!")

    @llm.call(
        provider=provider,
        model_id=model_id,
        tools=[secret_retrieval_tool],
    )
    async def call(passwords: list[str]) -> list[llm.Message]:
        return [
            llm.messages.system("Use parallel tool calling."),
            llm.messages.user(
                f"Please retrieve the secrets associated with each of these passwords: {','.join(passwords)}"
            ),
        ]

    with snapshot_test(snapshot) as snap:
        response = await call.stream(["mellon", "radiance"])
        await response.finish()

        assert len(response.tool_calls) == 2, (
            f"Expected response to have two tool calls: {response.pretty()}"
        )

        tool_outputs = await response.execute_tools()
        response = await response.resume(tool_outputs)
        await response.finish()

        snap.set_response(response)
        pretty = response.pretty()
        assert "Moria" in pretty, f"Expected 'Moria' to be in response: {pretty}"
        assert "Life before Death" in pretty, (
            f"Expected 'Life before Death' to be in response: {pretty}"
        )


@pytest.mark.parametrize("provider, model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_tools_async_stream_context(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test async streaming tool call with context."""

    @llm.tool
    async def secret_retrieval_tool(
        ctx: llm.Context[dict[str, str]], password: str
    ) -> str:
        """A tool that requires a password to retrieve a secret."""
        return ctx.deps.get(password, "Invalid password!")

    @llm.call(
        provider=provider,
        model_id=model_id,
        tools=[secret_retrieval_tool],
    )
    async def call(
        ctx: llm.Context[dict[str, str]], passwords: list[str]
    ) -> list[llm.Message]:
        return [
            llm.messages.system("Use parallel tool calling."),
            llm.messages.user(
                f"Please retrieve the secrets associated with each of these passwords: {','.join(passwords)}"
            ),
        ]

    with snapshot_test(snapshot) as snap:
        ctx = llm.Context(deps=PASSWORD_MAP)
        response = await call.stream(ctx, ["mellon", "radiance"])
        await response.finish()

        assert len(response.tool_calls) == 2, (
            f"Expected response to have two tool calls: {response.pretty()}"
        )

        tool_outputs = await response.execute_tools(ctx)
        response = await response.resume(ctx, tool_outputs)
        await response.finish()

        snap.set_response(response)
        pretty = response.pretty()
        assert "Moria" in pretty, f"Expected 'Moria' to be in response: {pretty}"
        assert "Life before Death" in pretty, (
            f"Expected 'Life before Death' to be in response: {pretty}"
        )



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Image,
    Text,
    URLImageSource,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider": "anthropic",
            "model_id": "claude-sonnet-4-0",
            "params": {},
            "finish_reason": None,
            "messages": [
                UserMessage(
                    content=[
                        Text(text="Describe the following image in one sentence"),
                        Image(
                            source=URLImageSource(
                                type="url_image_source",
                                url="https://en.wikipedia.org/static/images/icons/wikipedia.png",
                            )
                        ),
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="This is the iconic Wikipedia logo, featuring a spherical puzzle globe made up of white puzzle pieces with text in various languages and scripts from around the world."
                        )
                    ],
                    provider="anthropic",
                    model_id="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "citations": None,
                                "text": "This is the iconic Wikipedia logo, featuring a spherical puzzle globe made up of white puzzle pieces with text in various languages and scripts from around the world.",
                                "type": "text",
                            }
                        ],
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:completions",
            "model_id": "gpt-4o",
            "params": {},
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is 4200 + 42?")]),
                AssistantMessage(
                    content=[Text(text="\\(4200 + 42 = 4242\\).")],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message={
                        "content": "\\(4200 + 42 = 4242\\).",
                        "role": "assistant",
                        "annotations": [],
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:completions",
            "model_id": "gpt-4o",
            "params": {},
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is 4200 + 42?")]),
                AssistantMessage(
                    content=[Text(text="4200 + 42 is 4242.")],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message={
                        "content": "4200 + 42 is 4242.",
                        "role": "assistant",
                        "annotations": [],
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:completions",
            "model_id": "gpt-4o",
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is 4200 + 42?")]),
                AssistantMessage(
                    content=[Text(text="4200 + 42 equals 4242.")],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [],
            "n_chunks": 12,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider": "openai:completions",
            "model_id": "gpt-4o",
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is 4200 + 42?")]),
                AssistantMessage(
                    content=[Text(text="The sum of 4200 and 42 is 4242.")],
                    provider="openai:completions",
                    model_id="gpt-4o",
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [],
            "n_chunks": 16,
        }
    }
)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    Thought,
    ToolCall,
    ToolOutput,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider": "google",
            "model_id": "gemini-2.5-flash",
            "params": {"thinking": False},
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is the 100th fibonacci number?")]),
                AssistantMessage(
                    content=[
                        Thought(
                            thought='The user is asking for the 100th Fibonacci number. I have access to a function called "compute_fib" that can compute the nth Fibonacci number (1-indexed). The user has provided the specific value n=100, so I have all the required parameters to make the function call.'
                        ),
                        ToolCall(
                            id="toolu_01MnLnuRr9SYqk4Nt4zvBopq",
                            name="compute_fib",
                            args='{"n": 100}',
                        ),
                    ],
                    provider="anthropic",
                    model_id="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "signature": "ErQDCkYICRgCKkC1U5aLDccw6hfKtcWrClCTiZ6M12W9sfSIILD3F5gCggCOaELdyMLcXBM4QIU+YK9MZn5ZMu5v1Rx8ibkKyrbXEgxN0vJibplcGc6LU58aDI1emmpMRP8L7LzYaiIwmx+BjJtfz6CmWato1sgn/SInbtCLCoMAWXe51nSsFPtvUqF4qeOxliYQQjcxG6HOKpsC+PFTKFloh3uZJU18TX8i9GvZOe6UvoO10NdIZflTrRoBEZIPkGEh0X/8L2qD/ZyWKM+u2N0V7DzcEy+RA0Bczo+pzYmTJQo/xYrg+tOon3aboix8zWZvzolQatvxG9YIUhK8NtPTsFpOOFHmcCb2/Ap+piNfT3AWzKTy3h23Daaz6srxOuaHaW03t3vS4x24pMMrIleTbUcKFa7e+Uh7dSmrwZO5zBxo6RjauZvz5fV/Hz6LOYGHPC8uQ6fSAOgA1wQaRNvcdVHZVdvbWAxfhXoKCEOIRQB1KrqNa/sbo+ru6uXPeBsetHrRmLO8ZU3ZghgLpSrKKWvrdG01frZVS6+fgtdr3alTVsIhyIxXichJxe38IzvsTJyV2RgB",
                                "thinking": 'The user is asking for the 100th Fibonacci number. I have access to a function called "compute_fib" that can compute the nth Fibonacci number (1-indexed). The user has provided the specific value n=100, so I have all the required parameters to make the function call.',
                                "type": "thinking",
                            },
                            {
                                "id": "toolu_01MnLnuRr9SYqk4Nt4zvBopq",
                                "input": {"n": 100},
                                "name": "compute_fib",
                                "type": "tool_use",
                            },
                        ],
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="toolu_01MnLnuRr9SYqk4Nt4zvBopq",
                            name="compute_fib",
                            value="218922995834555169026",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="The 100th Fibonacci number is 218,922,995,834,555,169,026."
                        )
                    ],
                    provider="google",
                    model_id="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "video_metadata": None,
                                "thought": None,
                                "inline_data": None,
                                "file_data": None,
                                "thought_signature": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "function_call": None,
                                "function_response": None,
                                "text": "The 100th Fibonacci number is 218,922,995,834,555,169,026.",
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "compute_fib",
                    "description": "Compute the nth Fibonacci number (1-indexed).",
                    "parameters": """\
{
  "properties": {
    "n": {
      "title": "N",
      "type": "integer"
    }
  },
  "required": [
    "n"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
        }
    }
)



"""End-to-end tests for LLM call with thinking=True parameter."""

import pytest

from mirascope import llm
from tests.utils import (
    Snapshot,
    snapshot_test,
)

PROVIDER_MODEL_ID_PAIRS: list[tuple[llm.Provider, llm.ModelId]] = [
    ("openai:responses", "gpt-5"),
    ("anthropic", "claude-sonnet-4-0"),
    ("google", "gemini-2.5-flash"),
]

# This can't be easily answered without thinking
PROMPT = "How many primes below 400 contain 79 as a substring? Answer ONLY with the number, not sharing which primes they are."
# Check whether the model thinks it retained access to its reasoning process
RESUME_PROMPT = (
    "If you remember what the primes were, then share them, or say 'I don't remember.'"
)


@pytest.mark.parametrize("provider,model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
def test_call_with_thinking_true_sync(
    provider: llm.Provider,
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test synchronous call with thinking=True to verify reasoning content."""

    @llm.call(provider=provider, model_id=model_id, thinking=True)
    def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = call(PROMPT)
        with llm.model(provider=provider, model_id=model_id, thinking=False):
            response = response.resume(RESUME_PROMPT)
        snap.set_response(response)


@pytest.mark.parametrize("provider,model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
def test_call_with_thinking_true_stream(
    provider: llm.Provider,
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test streaming call with thinking=True to verify reasoning content."""

    @llm.call(provider=provider, model_id=model_id, thinking=True)
    def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = call.stream(PROMPT)
        response.finish()
        with llm.model(provider=provider, model_id=model_id, thinking=False):
            response = response.resume(RESUME_PROMPT)
        response.finish()
        snap.set_response(response)


@pytest.mark.parametrize("provider,model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_thinking_true_async(
    provider: llm.Provider,
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test asynchronous call with thinking=True to verify reasoning content."""

    @llm.call(provider=provider, model_id=model_id, thinking=True)
    async def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = await call(PROMPT)
        with llm.model(provider=provider, model_id=model_id, thinking=False):
            response = await response.resume(RESUME_PROMPT)
        snap.set_response(response)


@pytest.mark.parametrize("provider,model_id", PROVIDER_MODEL_ID_PAIRS)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_call_with_thinking_true_async_stream(
    provider: llm.Provider,
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test async streaming call with thinking=True to verify reasoning content."""

    @llm.call(provider=provider, model_id=model_id, thinking=True)
    async def call(query: str) -> str:
        return query

    with snapshot_test(snapshot, caplog) as snap:
        response = await call.stream(PROMPT)
        await response.finish()
        with llm.model(provider=provider, model_id=model_id, thinking=False):
            response = await response.resume(RESUME_PROMPT)
        await response.finish()
        snap.set_response(response)



"""Base class for StreamResponse and AsyncStreamResponse."""

from collections.abc import AsyncIterator, Iterator, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeAlias, TypeVar

from ..content import (
    AssistantContentChunk,
    AssistantContentPart,
    Text,
    TextChunk,
    TextEndChunk,
    TextStartChunk,
    Thought,
    ThoughtChunk,
    ThoughtEndChunk,
    ThoughtStartChunk,
    ToolCall,
    ToolCallChunk,
    ToolCallEndChunk,
    ToolCallStartChunk,
)
from ..formatting import Format, FormattableT, Partial
from ..messages import AssistantMessage, Message
from ..tools import FORMAT_TOOL_NAME, ToolkitT
from ..types import Jsonable
from .finish_reason import FinishReasonChunk
from .root_response import RootResponse
from .streams import (
    AsyncStream,
    AsyncTextStream,
    AsyncThoughtStream,
    AsyncToolCallStream,
    Stream,
    TextStream,
    ThoughtStream,
    ToolCallStream,
)

if TYPE_CHECKING:
    from ..clients import ModelId, Params, Provider


@dataclass(kw_only=True)
class RawStreamEventChunk:
    """A chunk containing a raw stream event from the underlying provider.

    Will be accumulated on `StreamResponse.raw` for debugging purposes.
    """

    type: Literal["raw_stream_event_chunk"] = "raw_stream_event_chunk"

    raw_stream_event: Any
    """The raw stream event from the underlying provider."""


@dataclass(kw_only=True)
class RawMessageChunk:
    """A chunk containing provider-specific raw message content that will be added to the `AssistantMessage`.

    This chunk contains a provider-specific representation of a piece of content that
    will be added to the `AssistantMessage` reconstructed by the containing stream.
    This content should be a Jsonable Python object for serialization purposes.

    The intention is that this content may be passed as-is back to the provider when the
    generated `AssistantMessage` is being reused in conversation.
    """

    type: Literal["raw_message_chunk"] = "raw_message_chunk"

    raw_message: Jsonable
    """The provider-specific raw content.
    
    Should be a Jsonable object.
    """


StreamResponseChunk: TypeAlias = (
    AssistantContentChunk | FinishReasonChunk | RawStreamEventChunk | RawMessageChunk
)

ChunkIterator: TypeAlias = Iterator[StreamResponseChunk]
"""Synchronous iterator yielding chunks with raw data."""

AsyncChunkIterator: TypeAlias = AsyncIterator[StreamResponseChunk]
"""Asynchronous iterator yielding chunks with raw data."""

ChunkIteratorT = TypeVar("ChunkIteratorT", bound=ChunkIterator | AsyncChunkIterator)


class BaseStreamResponse(
    RootResponse[ToolkitT, FormattableT],
    Generic[ChunkIteratorT, ToolkitT, FormattableT],
):
    """Base class underpinning StreamResponse and AsyncStreamResponse.

    Manages chunk handling logic for both.
    """

    raw_stream_events: Sequence[Any]
    """The raw stream event chunks from the LLM. Provider-specific."""

    chunks: Sequence[AssistantContentChunk]
    """All of the Mirascope chunks consumed from the stream."""

    content: Sequence[AssistantContentPart]
    """The content generated by the LLM.
    
    Content is updated in this array as it is consumed by the stream. Text content will 
    update with each text chunk (this will mutate the Text object that is returned 
    rather than creating a new one). Other content will be added once each part
    is fully streamed.
    """

    messages: list[Message]
    """The message history, including the most recent assistant message.

    The most recent assistant message will have all of the completed content that has 
    already been consumed from the stream. Text content will be included as each chunk 
    is processed; other content will be included only when its corresponding part is 
    completed (to avoid partial tool calls and the like). If no content has been 
    streamed, then the final assistant message will be present (to maintain turn order 
    expectations), but will be empty.
    """

    texts: Sequence[Text]
    """The text content in the generated response, if any.
    
    Text content updates with each text chunk as it streams. The `Text` objects are 
    mutated in place rather than creating new ones for each chunk.
    """

    tool_calls: Sequence[ToolCall]
    """The tools the LLM wants called on its behalf, if any.
    
    Tool calls are only added to this sequence once they have been fully streamed 
    to avoid partial tool calls in the response.
    """

    thoughts: Sequence[Thought]
    """The readable thoughts from the model's thinking process, if any.
    
    The thoughts may be direct output from the model thinking process, or may be a
    generated summary. (This depends on the provider; newer models tend to summarize.)

    Thoughts are added to the sequence as they are streamed. The `Thought` objects are
    mutated in place rather than creating new ones for each chunk.
    """

    consumed: bool = False
    """Whether the stream has been fully consumed.
    
    This is True after all chunks have been processed from the underlying iterator.
    When False, more content may be available by calling the stream methods.
    """

    def __init__(
        self,
        *,
        provider: "Provider",
        model_id: "ModelId",
        params: "Params",
        toolkit: ToolkitT,
        format: Format[FormattableT] | None = None,
        input_messages: Sequence[Message],
        chunk_iterator: ChunkIteratorT,
    ) -> None:
        """Initialize the BaseStreamResponse.

        Args:
            provider: The provider name (e.g. "anthropic", "openai:completions").
            model_id: The model identifier that generated the response.
            params: The params used to generate the response (or None).
            toolkit: Toolkit containing all the tools used to generate the response.
            format: The `Format` for the expected structured output format (or None).
            input_messages: The input messages that were sent to the LLM

        The BaseStreamResponse will process the tuples to build the chunks and raw lists
        as the stream is consumed.
        """

        self.provider = provider
        self.model_id = model_id
        self.params = params
        self.toolkit = toolkit
        self.format = format

        # Internal-only lists which we mutate (append) during chunk processing
        self._chunks: list[AssistantContentChunk] = []
        self._content: list[AssistantContentPart] = []
        self._texts: list[Text] = []
        self._thoughts: list[Thought] = []
        self._tool_calls: list[ToolCall] = []
        self._raw_stream_events: list[Any] = []
        self._last_raw_stream_event_chunk: Any | None = None

        # Externally-facing references typed as immutable Sequences
        self.chunks = self._chunks
        self.content = self._content
        self.texts = self._texts
        self.thoughts = self._thoughts
        self.tool_calls = self._tool_calls
        self.raw_stream_events = self._raw_stream_events

        self.finish_reason = None

        self._assistant_message = AssistantMessage(
            content=self._content,
            provider=provider,
            model_id=model_id,
            raw_message=None,
        )

        self.messages = list(input_messages) + [self._assistant_message]

        self._chunk_iterator = chunk_iterator
        self._current_content: Text | Thought | ToolCall | None = None

        self._processing_format_tool: bool = False

    def _transform_format_tool_chunks(
        self, chunk: AssistantContentChunk
    ) -> AssistantContentChunk:
        if chunk.type == "tool_call_start_chunk" and chunk.name.startswith(
            FORMAT_TOOL_NAME
        ):
            self._processing_format_tool = True
            return TextStartChunk()
        if self._processing_format_tool and chunk.type == "tool_call_chunk":
            return TextChunk(delta=chunk.delta)
        if self._processing_format_tool and chunk.type == "tool_call_end_chunk":
            self._processing_format_tool = False
            return TextEndChunk()
        return chunk

    def _handle_chunk(self, chunk: AssistantContentChunk) -> AssistantContentChunk:
        if self.finish_reason:
            raise RuntimeError(
                f"Stream already finished with reason: {self.finish_reason}"
            )
        chunk = self._transform_format_tool_chunks(chunk)

        if chunk.content_type == "text":
            self._handle_text_chunk(chunk)
        elif chunk.content_type == "tool_call":
            self._handle_tool_call_chunk(chunk)
        elif chunk.content_type == "thought":
            self._handle_thought_chunk(chunk)
        else:
            raise NotImplementedError

        self._chunks.append(chunk)
        return chunk

    def _handle_text_chunk(
        self, chunk: TextStartChunk | TextChunk | TextEndChunk
    ) -> None:
        if chunk.type == "text_start_chunk":
            if self._current_content:
                raise RuntimeError(
                    "Received text_start_chunk while processing another chunk"
                )
            self._current_content = Text(text="")
            # Text gets included in content even when unfinished.
            self._content.append(self._current_content)
            self._texts.append(self._current_content)

        elif chunk.type == "text_chunk":
            if self._current_content is None or self._current_content.type != "text":
                raise RuntimeError("Received text_chunk while not processing text.")
            self._current_content.text += chunk.delta

        elif chunk.type == "text_end_chunk":
            if self._current_content is None or self._current_content.type != "text":
                raise RuntimeError("Received text_end_chunk while not processing text.")
            self._current_content = None

    def _handle_thought_chunk(
        self, chunk: ThoughtStartChunk | ThoughtChunk | ThoughtEndChunk
    ) -> None:
        if chunk.type == "thought_start_chunk":
            if self._current_content:
                raise RuntimeError(
                    "Received thought_start_chunk while processing another chunk"
                )
            self._current_content = Thought(thought="")
            # Thoughts get included even when unfinished.
            self._content.append(self._current_content)
            self._thoughts.append(self._current_content)

        elif chunk.type == "thought_chunk":
            if self._current_content is None or self._current_content.type != "thought":
                raise RuntimeError(
                    "Received thought_chunk while not processing thought."
                )
            self._current_content.thought += chunk.delta

        elif chunk.type == "thought_end_chunk":
            if self._current_content is None or self._current_content.type != "thought":
                raise RuntimeError(
                    "Received thought_end_chunk while not processing thought."
                )
            self._current_content = None

    def _handle_tool_call_chunk(
        self, chunk: ToolCallStartChunk | ToolCallChunk | ToolCallEndChunk
    ) -> None:
        if chunk.type == "tool_call_start_chunk":
            if self._current_content:
                raise RuntimeError(
                    "Received tool_call_start_chunk while processing another chunk"
                )
            self._current_content = ToolCall(
                id=chunk.id,
                name=chunk.name,
                args="",
            )

        elif chunk.type == "tool_call_chunk":
            if (
                self._current_content is None
                or self._current_content.type != "tool_call"
            ):
                raise RuntimeError(
                    "Received tool_call_chunk while not processing tool call."
                )
            self._current_content.args += chunk.delta

        elif chunk.type == "tool_call_end_chunk":
            if (
                self._current_content is None
                or self._current_content.type != "tool_call"
            ):
                raise RuntimeError(
                    "Received tool_call_end_chunk while not processing tool call."
                )
            if not self._current_content.args:
                self._current_content.args = "{}"
            self._content.append(self._current_content)
            self._tool_calls.append(self._current_content)
            self._current_content = None

    def _pretty_chunk(self, chunk: AssistantContentChunk, spacer: str) -> str:
        match chunk.type:
            case "text_start_chunk":
                return spacer
            case "text_chunk":
                return chunk.delta
            case "tool_call_start_chunk":
                return spacer + f"**ToolCall ({chunk.name}):** "
            case "tool_call_chunk":
                return chunk.delta
            case "thought_start_chunk":
                return spacer + "**Thinking:**\n  "
            case "thought_chunk":
                return chunk.delta.replace("\n", "\n  ")  # Indent every line
            case _:
                return ""


class BaseSyncStreamResponse(BaseStreamResponse[ChunkIterator, ToolkitT, FormattableT]):
    """A base class for synchronous Stream Responses."""

    def streams(self) -> Iterator[Stream]:
        """Returns an iterator that yields streams for each content part in the response.

        Returns:
            Iterator[Stream]: Synchronous iterator yielding Stream objects

        Each content part in the response will correspond to one stream, which will yield
        chunks of content as they come in from the underlying LLM.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        As content is consumed, it is cached on the StreamResponse. If a new iterator
        is constructed via calling `streams()`, it will start by replaying the cached
        content from the response, and (if there is still more content to consume from
        the LLM), it will proceed to consume it once it has iterated through all the
        cached chunks.
        """
        chunk_iter = self.chunk_stream()

        for start_chunk in chunk_iter:
            # At the start of this loop, we always expect to find a start chunk. Then,
            # before proceeding, we will collect from the stream we create (in case the
            # user did not exhaust it), which ensures we will be expecting a start chunk
            # again on the next iteration
            match start_chunk.type:
                case "text_start_chunk":

                    def text_stream_iterator() -> Iterator[TextChunk]:
                        for chunk in chunk_iter:
                            if chunk.type == "text_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = TextStream(chunk_iterator=text_stream_iterator())
                    yield stream

                case "thought_start_chunk":

                    def thought_stream_iterator() -> Iterator[ThoughtChunk]:
                        for chunk in chunk_iter:
                            if chunk.type == "thought_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = ThoughtStream(chunk_iterator=thought_stream_iterator())
                    yield stream

                case "tool_call_start_chunk":
                    tool_id = start_chunk.id
                    tool_name = start_chunk.name

                    def tool_call_stream_iterator() -> Iterator[ToolCallChunk]:
                        for chunk in chunk_iter:
                            if chunk.type == "tool_call_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = ToolCallStream(
                        tool_id=tool_id,
                        tool_name=tool_name,
                        chunk_iterator=tool_call_stream_iterator(),
                    )
                    yield stream

                case _:  # pragma: no cover
                    raise RuntimeError(f"Expected start chunk, got: {start_chunk.type}")

            # Before continuing to the next stream, make sure the last stream is consumed
            # (If the user did not do so when we yielded it)
            stream.collect()

    def chunk_stream(
        self,
    ) -> Iterator[AssistantContentChunk]:
        """Returns an iterator that yields content chunks as they are received.

        Returns:
            Iterator[AssistantContentChunk]: Synchronous iterator yielding chunks

        This provides access to the Mirascope chunk data including start, delta, and end chunks
        for each content type (text, thought, tool_call). Unlike the streams() method
        that groups chunks by content part, this yields individual chunks as they arrive.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        As chunks are consumed, they are cached on the StreamResponse. If a new iterator
        is constructed via calling `chunk_stream()`, it will start by replaying the cached
        chunks from the response, and (if there is still more content to consume from
        the LLM), it will proceed to consume it once it has iterated through all the
        cached chunks.
        """
        for chunk in self.chunks:
            yield chunk

        if self.consumed:
            return

        for chunk in self._chunk_iterator:
            if chunk.type == "raw_stream_event_chunk":
                self._raw_stream_events.append(chunk.raw_stream_event)
            elif chunk.type == "raw_message_chunk":
                self._assistant_message.raw_message = chunk.raw_message
            elif chunk.type == "finish_reason_chunk":
                self.finish_reason = chunk.finish_reason
            else:
                yield self._handle_chunk(chunk)

        self.consumed = True

    def finish(self) -> None:
        """Finish streaming all of this response's content."""
        for _ in self.chunk_stream():
            pass

    def pretty_stream(self) -> Iterator[str]:
        """Stream a readable representation of the stream_response as text.

        Returns:
            Iterator[str]: Iterator yielding string chunks depicting the content

        Iterating through the pretty stream will populate the stream response by consuming
        the underlying iterator (if it hasn't been consumed already). Calling `.pretty_stream()`
        will always return a fresh iterator that begins from the start of the stream.

        If you concatenate the text from `.pretty_stream()`, it will be equivalent to the
        text generated by calling `.pretty()` (assuming the response was fully consumed
        at the time when you call `.pretty()`).
        """
        printed = False

        for chunk in self.chunk_stream():
            pretty = self._pretty_chunk(chunk, "\n\n" if printed else "")
            if pretty != "":
                printed = True
            yield pretty

        if not printed:
            yield "**[No Content]**"

    def structured_stream(
        self,
    ) -> Iterator[Partial[FormattableT]]:
        """Returns an iterator that yields partial structured objects as content streams.

        Returns:
            Iterator[Partial[FormatT]]: Synchronous iterator yielding partial structured objects

        This method yields Partial[FormatT] objects as the response content is streamed,
        allowing you to access partial structured data before the response is fully complete.
        Each yielded object represents the current state of the parsed structure with all
        fields optional.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.
        """
        raise NotImplementedError()


class BaseAsyncStreamResponse(
    BaseStreamResponse[AsyncChunkIterator, ToolkitT, FormattableT]
):
    """A base class for asynchronous Stream Responses."""

    async def streams(self) -> AsyncIterator[AsyncStream]:
        """Returns an async iterator that yields streams for each content part in the response.

        Returns:
            AsyncIterator[AsyncStream]: Async iterator yielding AsyncStream objects

        Each content part in the response will correspond to one stream, which will yield
        chunks of content as they come in from the underlying LLM.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        As content is consumed, it is cached on the AsyncStreamResponse. If a new iterator
        is constructed via calling `streams()`, it will start by replaying the cached
        content from the response, and (if there is still more content to consume from
        the LLM), it will proceed to consume it once it has iterated through all the
        cached chunks.
        """
        chunk_iter = self.chunk_stream()

        async for start_chunk in chunk_iter:
            # At the start of this loop, we always expect to find a start chunk. Then,
            # before proceeding, we will collect from the stream we create (in case the
            # user did not exhaust it), which ensures we will be expecting a start chunk
            # again on the next iteration
            match start_chunk.type:
                case "text_start_chunk":

                    async def text_stream_iterator() -> AsyncIterator[TextChunk]:
                        async for chunk in chunk_iter:
                            if chunk.type == "text_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = AsyncTextStream(chunk_iterator=text_stream_iterator())
                    yield stream

                case "thought_start_chunk":

                    async def thought_stream_iterator() -> AsyncIterator[ThoughtChunk]:
                        async for chunk in chunk_iter:
                            if chunk.type == "thought_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = AsyncThoughtStream(
                        chunk_iterator=thought_stream_iterator()
                    )
                    yield stream

                case "tool_call_start_chunk":
                    tool_id = start_chunk.id
                    tool_name = start_chunk.name

                    async def tool_call_stream_iterator() -> AsyncIterator[
                        ToolCallChunk
                    ]:
                        async for chunk in chunk_iter:
                            if chunk.type == "tool_call_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = AsyncToolCallStream(
                        tool_id=tool_id,
                        tool_name=tool_name,
                        chunk_iterator=tool_call_stream_iterator(),
                    )
                    yield stream

                case _:  # pragma: no cover
                    raise RuntimeError(f"Expected start chunk, got: {start_chunk.type}")

            # Before continuing to the next stream, make sure the last stream is consumed
            # (If the user did not do so when we yielded it)
            await stream.collect()

    async def chunk_stream(
        self,
    ) -> AsyncIterator[AssistantContentChunk]:
        """Returns an async iterator that yields content chunks as they are received.

        Returns:
            AsyncIterator[AssistantContentChunk]: Async iterator yielding chunks

        This provides access to the Mirascope chunk data including start, delta, and end chunks
        for each content type (text, thinking, tool_call). Unlike the streams() method
        that groups chunks by content part, this yields individual chunks as they arrive.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        As chunks are consumed, they are cached on the AsyncStreamResponse. If a new iterator
        is constructed via calling `chunk_stream()`, it will start by replaying the cached
        chunks from the response, and (if there is still more content to consume from
        the LLM), it will proceed to consume it once it has iterated through all the
        cached chunks.
        """

        for chunk in self.chunks:
            yield chunk

        if self.consumed:
            return

        async for chunk in self._chunk_iterator:
            if chunk.type == "raw_stream_event_chunk":
                self._raw_stream_events.append(chunk.raw_stream_event)
            elif chunk.type == "raw_message_chunk":
                self._assistant_message.raw_message = chunk.raw_message
            elif chunk.type == "finish_reason_chunk":
                self.finish_reason = chunk.finish_reason
            else:
                yield self._handle_chunk(chunk)

        self.consumed = True

    async def finish(self) -> None:
        """Finish streaming all of this response's content."""
        async for _ in self.chunk_stream():
            pass

    async def pretty_stream(self) -> AsyncIterator[str]:
        """Stream a readable representation of the stream_response as text.

        Returns:
            AsyncIterator[str]: Async iterator yielding string chunks depicting the content

        Iterating through the pretty stream will populate the stream response by consuming
        the underlying iterator (if it hasn't been consumed already). Calling `.pretty_stream()`
        will always return a fresh iterator that begins from the start of the stream.

        If you concatenate the text from `.pretty_stream()`, it will be equivalent to the
        text generated by calling `.pretty()` (assuming the response was fully consumed
        at the time when you call `.pretty()`).
        """
        printed = False

        async for chunk in self.chunk_stream():
            pretty = self._pretty_chunk(chunk, "\n\n" if printed else "")
            if pretty != "":
                printed = True
            yield pretty

        if not printed:
            yield "**[No Content]**"

    def structured_stream(
        self,
    ) -> AsyncIterator[Partial[FormattableT]]:
        """Returns an async iterator that yields partial structured objects as content streams.

        Returns:
            AsyncIterator[Partial[FormatT]]: Async iterator yielding partial structured objects

        This method yields Partial[FormatT] objects as the response content is streamed,
        allowing you to access partial structured data before the response is fully complete.
        Each yielded object represents the current state of the parsed structure with all
        fields optional.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.
        """
        raise NotImplementedError()



from mirascope import llm


@llm.call(provider="openai", model_id="gpt-5")
def recommend_book(genre: str):
    return f"Please recommend a book in {genre}."


def main():
    response: llm.StreamResponse = recommend_book.stream("fantasy")
    for chunk in response.pretty_stream():
        print(chunk, flush=True, end="")


main()



"""OpenAI Responses response decoding."""

from typing import Any, Literal

from openai import AsyncStream, Stream
from openai.types import responses as openai_types
from openai.types.responses.response_stream_event import ResponseStreamEvent

from .....content import (
    AssistantContentPart,
    Text,
    TextChunk,
    TextEndChunk,
    TextStartChunk,
    Thought,
    ThoughtChunk,
    ThoughtEndChunk,
    ThoughtStartChunk,
    ToolCall,
    ToolCallChunk,
    ToolCallEndChunk,
    ToolCallStartChunk,
)
from .....messages import AssistantMessage
from .....responses import (
    AsyncChunkIterator,
    ChunkIterator,
    FinishReason,
    FinishReasonChunk,
    RawMessageChunk,
    RawStreamEventChunk,
)
from ..model_ids import OpenAIResponsesModelId

INCOMPLETE_DETAILS_TO_FINISH_REASON = {
    "max_output_tokens": FinishReason.MAX_TOKENS,
    "content_filter": FinishReason.REFUSAL,
}


def _serialize_output_item(
    item: openai_types.ResponseOutputItem,
) -> dict[str, Any]:
    """Returns the item serialized as a dictionary."""
    return {key: value for key, value in item.model_dump().items() if value is not None}


def decode_response(
    response: openai_types.Response,
    model_id: OpenAIResponsesModelId,
) -> tuple[AssistantMessage, FinishReason | None]:
    """Convert OpenAI Responses Response to mirascope AssistantMessage."""
    parts: list[AssistantContentPart] = []
    finish_reason: FinishReason | None = None
    refused = False

    for output_item in response.output:
        if output_item.type == "message":
            for content in output_item.content:
                if content.type == "output_text":
                    parts.append(Text(text=content.text))
                elif content.type == "refusal":
                    parts.append(Text(text=content.refusal))
                    refused = True
        elif output_item.type == "function_call":
            parts.append(
                ToolCall(
                    id=output_item.call_id,
                    name=output_item.name,
                    args=output_item.arguments,
                )
            )
        elif output_item.type == "reasoning":
            for summary_part in output_item.summary:
                if summary_part.type == "summary_text":
                    parts.append(Thought(thought=summary_part.text))
            if output_item.content:  # pragma: no cover
                # TODO: Add test case covering this
                # (Likely their open-source models output reasoning_text rather than summaries)
                for reasoning_content in output_item.content:
                    if reasoning_content.type == "reasoning_text":
                        parts.append(Thought(thought=reasoning_content.text))

        else:
            raise NotImplementedError(f"Unsupported output item: {output_item.type}")

    if refused:
        finish_reason = FinishReason.REFUSAL
    elif details := response.incomplete_details:
        finish_reason = INCOMPLETE_DETAILS_TO_FINISH_REASON.get(details.reason or "")

    assistant_message = AssistantMessage(
        content=parts,
        provider="openai:responses",
        model_id=model_id,
        raw_message=[
            _serialize_output_item(output_item) for output_item in response.output
        ],
    )

    return assistant_message, finish_reason


class _OpenAIResponsesChunkProcessor:
    """Processes OpenAI Responses streaming events and maintains state across chunks."""

    def __init__(self) -> None:
        self.current_content_type: Literal["text", "tool_call", "thought"] | None = None
        self.refusal_encountered = False

    def process_chunk(self, event: ResponseStreamEvent) -> ChunkIterator:
        """Process a single OpenAI Responses stream event and yield the appropriate content chunks."""
        yield RawStreamEventChunk(raw_stream_event=event)

        if hasattr(event, "type"):
            if event.type == "response.output_text.delta":
                if not self.current_content_type:
                    yield TextStartChunk()
                    self.current_content_type = "text"
                yield TextChunk(delta=event.delta)
            elif event.type == "response.output_text.done":
                yield TextEndChunk()
                self.current_content_type = None
            if event.type == "response.refusal.delta":
                if not self.current_content_type:
                    yield TextStartChunk()
                    self.current_content_type = "text"
                yield TextChunk(delta=event.delta)
            elif event.type == "response.refusal.done":
                yield TextEndChunk()
                self.refusal_encountered = True
                self.current_content_type = None
            elif event.type == "response.output_item.added":
                item = event.item
                if item.type == "function_call":
                    self.current_tool_call_id = item.call_id
                    self.current_tool_call_name = item.name
                    yield ToolCallStartChunk(
                        id=item.call_id,
                        name=item.name,
                    )
                    self.current_content_type = "tool_call"
            elif event.type == "response.function_call_arguments.delta":
                yield ToolCallChunk(delta=event.delta)
            elif event.type == "response.function_call_arguments.done":
                yield ToolCallEndChunk()
                self.current_content_type = None
            elif (
                event.type == "response.reasoning_text.delta"
                or event.type == "response.reasoning_summary_text.delta"
            ):
                if not self.current_content_type:
                    yield ThoughtStartChunk()
                    self.current_content_type = "thought"
                yield ThoughtChunk(delta=event.delta)
            elif (
                event.type == "response.reasoning_summary_text.done"
                or event.type == "response.reasoning_text.done"
            ):
                yield ThoughtEndChunk()
                self.current_content_type = None
            elif event.type == "response.incomplete":
                details = event.response.incomplete_details
                reason = (details and details.reason) or ""
                finish_reason = INCOMPLETE_DETAILS_TO_FINISH_REASON.get(reason)
                if finish_reason:
                    yield FinishReasonChunk(finish_reason=finish_reason)
            elif event.type == "response.completed":
                yield RawMessageChunk(
                    raw_message=[
                        _serialize_output_item(item) for item in event.response.output
                    ]
                )
                if self.refusal_encountered:
                    yield FinishReasonChunk(finish_reason=FinishReason.REFUSAL)


def decode_stream(
    openai_stream: Stream[ResponseStreamEvent],
) -> ChunkIterator:
    """Returns a ChunkIterator converted from an OpenAI Stream[ResponseStreamEvent]"""
    processor = _OpenAIResponsesChunkProcessor()
    for event in openai_stream:
        yield from processor.process_chunk(event)


async def decode_async_stream(
    openai_stream: AsyncStream[ResponseStreamEvent],
) -> AsyncChunkIterator:
    """Returns an AsyncChunkIterator converted from an OpenAI AsyncStream[ResponseStreamEvent]"""
    processor = _OpenAIResponsesChunkProcessor()
    async for event in openai_stream:
        for item in processor.process_chunk(event):
            yield item



"""Tests for Audio content class."""

import tempfile
from pathlib import Path

import httpx
import pytest
from pytest_httpserver import HTTPServer

from mirascope.llm.content.audio import MAX_AUDIO_SIZE, Audio, Base64AudioSource


@pytest.fixture
def audio_data() -> dict[str, bytes]:
    """Provide sample audio data for different formats (minimum 12 bytes each)."""
    return {
        "wav": b"RIFF\x00\x00\x00\x00WAVE",
        "mp3_id3": b"ID3\x03\x00\x00\x00\x00\x00\x00\x00\x00",
        "mp3_frame": b"\xff\xfb\x90\x00\x00\x00\x00\x00\x00\x00\x00\x00",
        "aiff": b"FORM\x00\x00\x00\x00AIFF",
        "aac_f1": b"\xff\xf1\x50\x80\x00\x00\x00\x00\x00\x00\x00\x00",
        "aac_f9": b"\xff\xf9\x50\x80\x00\x00\x00\x00\x00\x00\x00\x00",
        "ogg": b"OggS\x00\x02\x00\x00\x00\x00\x00\x00",
        "flac": b"fLaC\x00\x00\x00\x22\x00\x00\x00\x00",
        "unsupported": b"random unsupported data",
    }


AUDIO_FORMAT_TESTS = [
    ("wav", "audio/wav"),
    ("mp3_id3", "audio/mp3"),
    ("mp3_frame", "audio/mp3"),
    ("aiff", "audio/aiff"),
    ("aac_f1", "audio/aac"),
    ("aac_f9", "audio/aac"),
    ("ogg", "audio/ogg"),
    ("flac", "audio/flac"),
]


class TestAudioDownload:
    """Tests for Audio.download class method."""

    @pytest.mark.parametrize("format_name,expected_mime", AUDIO_FORMAT_TESTS)
    def test_download_detects_mime_type_from_magic_bytes(
        self,
        httpserver: HTTPServer,
        audio_data: dict,
        format_name: str,
        expected_mime: str,
    ) -> None:
        """Test that download() detects MIME type from magic bytes."""
        httpserver.expect_request("/audio").respond_with_data(audio_data[format_name])
        url = httpserver.url_for("/audio")

        audio = Audio.download(url)

        assert isinstance(audio.source, Base64AudioSource)
        assert audio.source.type == "base64_audio_source"
        assert audio.source.mime_type == expected_mime
        assert len(audio.source.data) > 0

    def test_download_follows_redirects(
        self, httpserver: HTTPServer, audio_data: dict
    ) -> None:
        """Test that download follows redirects."""
        httpserver.expect_request("/redirect").respond_with_data(
            "", status=302, headers={"Location": "/audio"}
        )
        httpserver.expect_request("/audio").respond_with_data(audio_data["wav"])
        url = httpserver.url_for("/redirect")

        audio = Audio.download(url)

        assert isinstance(audio.source, Base64AudioSource)
        assert audio.source.mime_type == "audio/wav"

    def test_download_raises_on_unsupported_format(
        self, httpserver: HTTPServer, audio_data: dict
    ) -> None:
        """Test that download raises ValueError for unsupported audio format."""
        httpserver.expect_request("/audio").respond_with_data(audio_data["unsupported"])
        url = httpserver.url_for("/audio")

        with pytest.raises(ValueError, match="Unsupported audio type"):
            Audio.download(url)

    def test_download_raises_on_http_error(self, httpserver: HTTPServer) -> None:
        """Test that download raises on HTTP errors."""
        httpserver.expect_request("/audio").respond_with_data("", status=404)
        url = httpserver.url_for("/audio")

        with pytest.raises(httpx.HTTPStatusError):
            Audio.download(url)

    def test_download_enforces_size_limit(self, httpserver: HTTPServer) -> None:
        """Test that download enforces the size limit."""
        large_data = b"RIFF\x00\x00\x00\x00WAVE" + b"x" * (MAX_AUDIO_SIZE + 1)
        httpserver.expect_request("/audio").respond_with_data(large_data)
        url = httpserver.url_for("/audio")

        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            Audio.download(url)

    def test_download_respects_custom_size_limit(
        self, httpserver: HTTPServer, audio_data: dict
    ) -> None:
        """Test that download respects custom max_size parameter."""
        httpserver.expect_request("/audio").respond_with_data(audio_data["wav"])
        url = httpserver.url_for("/audio")

        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            Audio.download(url, max_size=10)


class TestAudioDownloadAsync:
    """Tests for Audio.download_async class method."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "format_name,expected_mime",
        [
            ("wav", "audio/wav"),
            ("mp3_id3", "audio/mp3"),
            ("aiff", "audio/aiff"),
            ("ogg", "audio/ogg"),
        ],
    )
    async def test_download_async_detects_mime_type(
        self,
        httpserver: HTTPServer,
        audio_data: dict,
        format_name: str,
        expected_mime: str,
    ) -> None:
        """Test that download_async() detects MIME type from magic bytes."""
        httpserver.expect_request("/audio").respond_with_data(audio_data[format_name])
        url = httpserver.url_for("/audio")

        audio = await Audio.download_async(url)

        assert isinstance(audio.source, Base64AudioSource)
        assert audio.source.mime_type == expected_mime

    @pytest.mark.asyncio
    async def test_download_async_enforces_size_limit(
        self, httpserver: HTTPServer
    ) -> None:
        """Test that download_async enforces the size limit."""
        large_data = b"RIFF\x00\x00\x00\x00WAVE" + b"x" * (MAX_AUDIO_SIZE + 1)
        httpserver.expect_request("/audio").respond_with_data(large_data)
        url = httpserver.url_for("/audio")

        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            await Audio.download_async(url)

    @pytest.mark.asyncio
    async def test_download_async_raises_on_unsupported_format(
        self, httpserver: HTTPServer, audio_data: dict
    ) -> None:
        """Test that download_async raises ValueError for unsupported format."""
        httpserver.expect_request("/audio").respond_with_data(audio_data["unsupported"])
        url = httpserver.url_for("/audio")

        with pytest.raises(ValueError, match="Unsupported audio type"):
            await Audio.download_async(url)


class TestAudioFromFile:
    """Tests for Audio.from_file class method."""

    @pytest.mark.parametrize("data_key,expected_mime", AUDIO_FORMAT_TESTS)
    def test_from_file_detects_mime_type(
        self, audio_data: dict, data_key: str, expected_mime: str
    ) -> None:
        """Test that from_file detects MIME type from magic bytes (not extension)."""
        with tempfile.NamedTemporaryFile(suffix=".unknown", delete=False) as f:
            f.write(audio_data[data_key])
            temp_path = f.name

        try:
            audio = Audio.from_file(temp_path)

            assert isinstance(audio.source, Base64AudioSource)
            assert audio.source.mime_type == expected_mime
        finally:
            Path(temp_path).unlink()

    def test_from_file_raises_on_unsupported_format(self, audio_data: dict) -> None:
        """Test that from_file raises ValueError for unsupported format."""
        with tempfile.NamedTemporaryFile(suffix=".bin", delete=False) as f:
            f.write(audio_data["unsupported"])
            temp_path = f.name

        try:
            with pytest.raises(ValueError, match="Unsupported audio type"):
                Audio.from_file(temp_path)
        finally:
            Path(temp_path).unlink()

    def test_from_file_not_found(self) -> None:
        """Test FileNotFoundError is raised for non-existent file."""
        with pytest.raises(FileNotFoundError):
            Audio.from_file("/nonexistent/path/to/file.mp3")

    def test_from_file_enforces_size_limit(self, audio_data: dict) -> None:
        """Test that from_file enforces the size limit."""
        large_data = b"RIFF\x00\x00\x00\x00WAVE" + b"x" * (MAX_AUDIO_SIZE + 1)
        with tempfile.NamedTemporaryFile(suffix=".wav", delete=False) as f:
            f.write(large_data)
            temp_path = f.name

        try:
            with pytest.raises(ValueError, match="exceeds maximum allowed size"):
                Audio.from_file(temp_path)
        finally:
            Path(temp_path).unlink()

    def test_from_file_respects_custom_size_limit(self, audio_data: dict) -> None:
        """Test that from_file respects custom max_size parameter."""
        with tempfile.NamedTemporaryFile(suffix=".wav", delete=False) as f:
            f.write(audio_data["wav"])
            temp_path = f.name

        try:
            with pytest.raises(ValueError, match="exceeds maximum allowed size"):
                Audio.from_file(temp_path, max_size=10)
        finally:
            Path(temp_path).unlink()


class TestAudioFromBytes:
    """Tests for Audio.from_bytes class method."""

    @pytest.mark.parametrize("data_key,expected_mime", AUDIO_FORMAT_TESTS)
    def test_from_bytes_detects_mime_type(
        self, audio_data: dict, data_key: str, expected_mime: str
    ) -> None:
        """Test that from_bytes detects MIME type from magic bytes."""
        audio = Audio.from_bytes(audio_data[data_key])

        assert isinstance(audio.source, Base64AudioSource)
        assert audio.source.mime_type == expected_mime

    def test_from_bytes_raises_on_unsupported_format(self, audio_data: dict) -> None:
        """Test that from_bytes raises ValueError for unsupported format."""
        with pytest.raises(ValueError, match="Unsupported audio type"):
            Audio.from_bytes(audio_data["unsupported"])

    def test_from_bytes_raises_on_data_too_small(self) -> None:
        """Test that from_bytes raises ValueError for data that's too small."""
        with pytest.raises(ValueError, match="Audio data too small to determine type"):
            Audio.from_bytes(b"short")

    def test_from_bytes_enforces_size_limit(self) -> None:
        """Test that from_bytes enforces the size limit."""
        large_data = b"RIFF\x00\x00\x00\x00WAVE" + b"x" * (MAX_AUDIO_SIZE + 1)

        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            Audio.from_bytes(large_data)

    def test_from_bytes_respects_custom_size_limit(self, audio_data: dict) -> None:
        """Test that from_bytes respects custom max_size parameter."""
        with pytest.raises(ValueError, match="exceeds maximum allowed size"):
            Audio.from_bytes(audio_data["wav"], max_size=10)



"""OpenAI Responses API client implementation."""

import os
from collections.abc import Sequence
from contextvars import ContextVar
from functools import lru_cache
from typing import overload
from typing_extensions import Unpack

from openai import AsyncOpenAI, OpenAI

from ....context import Context, DepsT
from ....formatting import Format, FormattableT
from ....messages import Message
from ....responses import (
    AsyncContextResponse,
    AsyncContextStreamResponse,
    AsyncResponse,
    AsyncStreamResponse,
    ContextResponse,
    ContextStreamResponse,
    Response,
    StreamResponse,
)
from ....tools import (
    AsyncContextTool,
    AsyncContextToolkit,
    AsyncTool,
    AsyncToolkit,
    ContextTool,
    ContextToolkit,
    Tool,
    Toolkit,
)
from ...base import BaseClient, Params
from . import _utils
from .model_ids import OpenAIResponsesModelId

OPENAI_RESPONSES_CLIENT_CONTEXT: ContextVar["OpenAIResponsesClient | None"] = (
    ContextVar("OPENAI_RESPONSES_CLIENT_CONTEXT", default=None)
)


@lru_cache(maxsize=256)
def _openai_responses_singleton(
    api_key: str | None, base_url: str | None
) -> "OpenAIResponsesClient":
    """Return a cached `OpenAIResponsesClient` instance for the given parameters."""
    return OpenAIResponsesClient(api_key=api_key, base_url=base_url)


def client(
    *, api_key: str | None = None, base_url: str | None = None
) -> "OpenAIResponsesClient":
    """Return an `OpenAIResponsesClient`."""
    api_key = api_key or os.getenv("OPENAI_API_KEY")
    base_url = base_url or os.getenv("OPENAI_BASE_URL")
    return _openai_responses_singleton(api_key, base_url)


def get_client() -> "OpenAIResponsesClient":
    """Get the current `OpenAIResponsesClient` from context."""
    current_client = OPENAI_RESPONSES_CLIENT_CONTEXT.get()
    if current_client is None:
        current_client = client()
        OPENAI_RESPONSES_CLIENT_CONTEXT.set(current_client)
    return current_client


class OpenAIResponsesClient(BaseClient[OpenAIResponsesModelId, OpenAI]):
    """The client for the OpenAI Responses API."""

    @property
    def _context_var(self) -> ContextVar["OpenAIResponsesClient | None"]:
        return OPENAI_RESPONSES_CLIENT_CONTEXT

    def __init__(
        self, *, api_key: str | None = None, base_url: str | None = None
    ) -> None:
        """Initialize the OpenAI Responses client."""
        self.client = OpenAI(api_key=api_key, base_url=base_url)
        self.async_client = AsyncOpenAI(api_key=api_key, base_url=base_url)

    @overload
    def call(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> Response:
        """Generate an `llm.Response` without a response format."""
        ...

    @overload
    def call(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> Response[FormattableT]:
        """Generate an `llm.Response` with a response format."""
        ...

    @overload
    def call(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> Response | Response[FormattableT]:
        """Generate an `llm.Response` with optional response format."""
        ...

    def call(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> Response | Response[FormattableT]:
        """Generate an `llm.Response` by synchronously calling the OpenAI Responses API.

        Args:
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.Response` object containing the LLM-generated content.
        """
        messages, format, kwargs = _utils.encode_request(
            model_id=model_id,
            messages=messages,
            tools=tools,
            format=format,
            params=params,
        )

        openai_response = self.client.responses.create(**kwargs)

        assistant_message, finish_reason = _utils.decode_response(
            openai_response, model_id
        )

        return Response(
            raw=openai_response,
            provider="openai:responses",
            model_id=model_id,
            params=params,
            tools=tools,
            input_messages=messages,
            assistant_message=assistant_message,
            finish_reason=finish_reason,
            format=format,
        )

    @overload
    async def call_async(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> AsyncResponse:
        """Generate an `llm.AsyncResponse` without a response format."""
        ...

    @overload
    async def call_async(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> AsyncResponse[FormattableT]:
        """Generate an `llm.AsyncResponse` with a response format."""
        ...

    @overload
    async def call_async(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generate an `llm.AsyncResponse` with optional response format."""
        ...

    async def call_async(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generate an `llm.AsyncResponse` by asynchronously calling the OpenAI Responses API.

        Args:
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.AsyncResponse` object containing the LLM-generated content.
        """
        messages, format, kwargs = _utils.encode_request(
            model_id=model_id,
            messages=messages,
            tools=tools,
            format=format,
            params=params,
        )

        openai_response = await self.async_client.responses.create(**kwargs)

        assistant_message, finish_reason = _utils.decode_response(
            openai_response, model_id
        )

        return AsyncResponse(
            raw=openai_response,
            provider="openai:responses",
            model_id=model_id,
            params=params,
            tools=tools,
            input_messages=messages,
            assistant_message=assistant_message,
            finish_reason=finish_reason,
            format=format,
        )

    @overload
    def stream(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> StreamResponse:
        """Generate a `llm.StreamResponse` without a response format."""
        ...

    @overload
    def stream(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> StreamResponse[FormattableT]:
        """Generate a `llm.StreamResponse` with a response format."""
        ...

    @overload
    def stream(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Generate a `llm.StreamResponse` with optional response format."""
        ...

    def stream(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Generate a `llm.StreamResponse` by synchronously streaming from the OpenAI Responses API.

        Args:
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A `llm.StreamResponse` object containing the LLM-generated content stream.
        """
        messages, format, kwargs = _utils.encode_request(
            model_id=model_id,
            messages=messages,
            tools=tools,
            format=format,
            params=params,
        )

        openai_stream = self.client.responses.create(
            **kwargs,
            stream=True,
        )

        chunk_iterator = _utils.decode_stream(
            openai_stream,
        )

        return StreamResponse(
            provider="openai:responses",
            model_id=model_id,
            params=params,
            tools=tools,
            input_messages=messages,
            chunk_iterator=chunk_iterator,
            format=format,
        )

    @overload
    async def stream_async(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse:
        """Generate a `llm.AsyncStreamResponse` without a response format."""
        ...

    @overload
    async def stream_async(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> AsyncStreamResponse[FormattableT]:
        """Generate a `llm.AsyncStreamResponse` with a response format."""
        ...

    @overload
    async def stream_async(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Generate a `llm.AsyncStreamResponse` with optional response format."""
        ...

    async def stream_async(
        self,
        *,
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Generate a `llm.AsyncStreamResponse` by asynchronously streaming from the OpenAI Responses API.

        Args:
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A `llm.AsyncStreamResponse` object containing the LLM-generated content stream.
        """
        messages, format, kwargs = _utils.encode_request(
            model_id=model_id,
            messages=messages,
            tools=tools,
            format=format,
            params=params,
        )

        openai_stream = await self.async_client.responses.create(
            **kwargs,
            stream=True,
        )

        chunk_iterator = _utils.decode_async_stream(
            openai_stream,
        )

        return AsyncStreamResponse(
            provider="openai:responses",
            model_id=model_id,
            params=params,
            tools=tools,
            input_messages=messages,
            chunk_iterator=chunk_iterator,
            format=format,
        )

    @overload
    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT]:
        """Generate a `llm.ContextResponse` without a response format."""
        ...

    @overload
    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, FormattableT]:
        """Generate a `llm.ContextResponse` with a response format."""
        ...

    @overload
    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT] | ContextResponse[DepsT, FormattableT]:
        """Generate a `llm.ContextResponse` with optional response format."""
        ...

    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT] | ContextResponse[DepsT, FormattableT]:
        """Generate a `llm.ContextResponse` by synchronously calling the OpenAI Responses API with context.

        Args:
            ctx: The context object containing dependencies.
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A `llm.ContextResponse` object containing the LLM-generated content and context.
        """
        messages, format, kwargs = _utils.encode_request(
            model_id=model_id,
            messages=messages,
            tools=tools,
            format=format,
            params=params,
        )

        openai_response = self.client.responses.create(**kwargs)

        assistant_message, finish_reason = _utils.decode_response(
            openai_response, model_id
        )

        return ContextResponse(
            raw=openai_response,
            provider="openai:responses",
            model_id=model_id,
            params=params,
            tools=tools,
            input_messages=messages,
            assistant_message=assistant_message,
            finish_reason=finish_reason,
            format=format,
        )

    @overload
    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT]:
        """Generate a `llm.AsyncContextResponse` without a response format."""
        ...

    @overload
    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, FormattableT]:
        """Generate a `llm.AsyncContextResponse` with a response format."""
        ...

    @overload
    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT] | AsyncContextResponse[DepsT, FormattableT]:
        """Generate a `llm.AsyncContextResponse` with optional response format."""
        ...

    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT] | AsyncContextResponse[DepsT, FormattableT]:
        """Generate a `llm.AsyncContextResponse` by asynchronously calling the OpenAI Responses API with context.

        Args:
            ctx: The context object containing dependencies.
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A `llm.AsyncContextResponse` object containing the LLM-generated content and context.
        """
        messages, format, kwargs = _utils.encode_request(
            model_id=model_id,
            messages=messages,
            tools=tools,
            format=format,
            params=params,
        )

        openai_response = await self.async_client.responses.create(**kwargs)

        assistant_message, finish_reason = _utils.decode_response(
            openai_response, model_id
        )

        return AsyncContextResponse(
            raw=openai_response,
            provider="openai:responses",
            model_id=model_id,
            params=params,
            tools=tools,
            input_messages=messages,
            assistant_message=assistant_message,
            finish_reason=finish_reason,
            format=format,
        )

    @overload
    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> ContextStreamResponse[DepsT]:
        """Generate a `llm.ContextStreamResponse` without a response format."""
        ...

    @overload
    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> ContextStreamResponse[DepsT, FormattableT]:
        """Generate a `llm.ContextStreamResponse` with a response format."""
        ...

    @overload
    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> ContextStreamResponse[DepsT] | ContextStreamResponse[DepsT, FormattableT]:
        """Generate a `llm.ContextStreamResponse` with optional response format."""
        ...

    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> ContextStreamResponse[DepsT] | ContextStreamResponse[DepsT, FormattableT]:
        """Generate a `llm.ContextStreamResponse` by synchronously streaming from the OpenAI Responses API with context.

        Args:
            ctx: The context object containing dependencies.
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A `llm.ContextStreamResponse` object containing the LLM-generated content stream and context.
        """
        messages, format, kwargs = _utils.encode_request(
            model_id=model_id,
            messages=messages,
            tools=tools,
            format=format,
            params=params,
        )

        openai_stream = self.client.responses.create(
            **kwargs,
            stream=True,
        )

        chunk_iterator = _utils.decode_stream(
            openai_stream,
        )

        return ContextStreamResponse(
            provider="openai:responses",
            model_id=model_id,
            params=params,
            tools=tools,
            input_messages=messages,
            chunk_iterator=chunk_iterator,
            format=format,
        )

    @overload
    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> AsyncContextStreamResponse[DepsT]:
        """Generate a `llm.AsyncContextStreamResponse` without a response format."""
        ...

    @overload
    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> AsyncContextStreamResponse[DepsT, FormattableT]:
        """Generate a `llm.AsyncContextStreamResponse` with a response format."""
        ...

    @overload
    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> (
        AsyncContextStreamResponse[DepsT]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate a `llm.AsyncContextStreamResponse` with optional response format."""
        ...

    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: OpenAIResponsesModelId,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> (
        AsyncContextStreamResponse[DepsT]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate a `llm.AsyncContextStreamResponse` by asynchronously streaming from the OpenAI Responses API with context.

        Args:
            ctx: The context object containing dependencies.
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A `llm.AsyncContextStreamResponse` object containing the LLM-generated content stream and context.
        """
        messages, format, kwargs = _utils.encode_request(
            model_id=model_id,
            messages=messages,
            tools=tools,
            format=format,
            params=params,
        )

        openai_stream = await self.async_client.responses.create(
            **kwargs,
            stream=True,
        )

        chunk_iterator = _utils.decode_async_stream(
            openai_stream,
        )

        return AsyncContextStreamResponse(
            provider="openai:responses",
            model_id=model_id,
            params=params,
            tools=tools,
            input_messages=messages,
            chunk_iterator=chunk_iterator,
            format=format,
        )



"""End-to-end tests for behavior when reaching max_tokens limits"""

import pytest

from mirascope import llm
from tests.e2e.conftest import PROVIDER_MODEL_ID_PAIRS
from tests.utils import (
    Snapshot,
    snapshot_test,
)


@pytest.mark.parametrize(
    "provider,model_id",
    PROVIDER_MODEL_ID_PAIRS,
)
@pytest.mark.vcr
def test_max_tokens_sync(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test synchronous call with token limits."""

    @llm.call(provider=provider, model_id=model_id, max_tokens=50)
    def list_states() -> str:
        return "List all U.S. states."

    with snapshot_test(snapshot) as snap:
        response = list_states()
        snap.set_response(response)
        assert response.finish_reason == llm.FinishReason.MAX_TOKENS


@pytest.mark.parametrize(
    "provider,model_id",
    PROVIDER_MODEL_ID_PAIRS,
)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_max_tokens_async(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test asynchronous call with token limits."""

    @llm.call(provider=provider, model_id=model_id, max_tokens=50)
    async def list_states() -> str:
        return "List all U.S. states."

    with snapshot_test(snapshot) as snap:
        response = await list_states()
        snap.set_response(response)
        assert response.finish_reason == llm.FinishReason.MAX_TOKENS


@pytest.mark.parametrize(
    "provider,model_id",
    PROVIDER_MODEL_ID_PAIRS,
)
@pytest.mark.vcr
def test_max_tokens_stream(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test streaming call with token limits."""

    @llm.call(provider=provider, model_id=model_id, max_tokens=50)
    def list_states() -> str:
        return "List all U.S. states."

    with snapshot_test(snapshot) as snap:
        response = list_states.stream()
        response.finish()
        snap.set_response(response)
        assert response.finish_reason == llm.FinishReason.MAX_TOKENS


@pytest.mark.parametrize(
    "provider,model_id",
    PROVIDER_MODEL_ID_PAIRS,
)
@pytest.mark.vcr
@pytest.mark.asyncio
async def test_max_tokens_async_stream(
    provider: llm.Provider, model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test async streaming call with token limits."""

    @llm.call(provider=provider, model_id=model_id, max_tokens=50)
    async def list_states() -> str:
        return "List all U.S. states."

    with snapshot_test(snapshot) as snap:
        response = await list_states.stream()
        await response.finish()
        snap.set_response(response)
        assert response.finish_reason == llm.FinishReason.MAX_TOKENS



"""Google message encoding and request preparation."""

import base64
import json
from collections.abc import Sequence
from functools import lru_cache
from typing import Any, cast

from google.genai import types as genai_types

from ....content import ContentPart
from ....exceptions import FeatureNotSupportedError
from ....formatting import (
    Format,
    FormattableT,
    _utils as _formatting_utils,
    resolve_format,
)
from ....messages import AssistantMessage, Message, UserMessage
from ....tools import FORMAT_TOOL_NAME, BaseToolkit, ToolSchema
from ...base import BaseKwargs, Params, _utils as _base_utils
from ..model_ids import GoogleModelId

UNKNOWN_TOOL_ID = "google_unknown_tool_id"


class GoogleKwargs(BaseKwargs, genai_types.GenerateContentConfigDict):
    """Google's `GenerateContentConfigDict` typed dict, subclassing BaseKwargs for type safety."""


def _resolve_refs(
    schema: dict[str, Any], defs: dict[str, Any] | None = None
) -> dict[str, Any]:
    """Recursively resolve $ref references in a JSON Schema object."""
    if isinstance(schema, dict):
        if "$ref" in schema:
            ref_path = schema["$ref"]
            if ref_path.startswith("#/$defs/"):
                ref_name = ref_path.split("/")[-1]
                if defs and ref_name in defs:
                    return _resolve_refs(defs[ref_name], defs)
            return schema  # pragma: no cover
        else:
            return {k: _resolve_refs(v, defs) for k, v in schema.items()}
    elif isinstance(schema, list):
        return [_resolve_refs(item, defs) for item in schema]
    else:
        return schema


def _encode_content(
    content: Sequence[ContentPart], encode_thoughts: bool
) -> list[genai_types.PartDict]:
    """Returns a list of google `PartDicts` converted from a sequence of Mirascope `ContentPart`s"""
    result = []

    for part in content:
        if part.type == "text":
            result.append(genai_types.PartDict(text=part.text))
        elif part.type == "image":
            if part.source.type == "base64_image_source":
                result.append(
                    genai_types.PartDict(
                        inline_data=genai_types.BlobDict(
                            data=base64.b64decode(part.source.data),
                            mime_type=part.source.mime_type,
                        )
                    )
                )
            elif part.source.type == "url_image_source":
                raise FeatureNotSupportedError(
                    "url_image_source",
                    "google",
                    message="Google does not support URL-referenced images. Try `llm.Image.download(...) or `llm.Image.download_async(...)`",
                )
        elif part.type == "audio":
            if part.source.type == "base64_audio_source":
                result.append(
                    genai_types.PartDict(
                        inline_data=genai_types.BlobDict(
                            data=base64.b64decode(part.source.data),
                            mime_type=part.source.mime_type,
                        )
                    )
                )
        elif part.type == "tool_call":
            result.append(
                genai_types.PartDict(
                    function_call=genai_types.FunctionCallDict(
                        name=part.name,
                        args=json.loads(part.args),
                        id=part.id if part.id != UNKNOWN_TOOL_ID else None,
                    )
                )
            )
        elif part.type == "tool_output":
            result.append(
                genai_types.PartDict(
                    function_response=genai_types.FunctionResponseDict(
                        id=part.id if part.id != UNKNOWN_TOOL_ID else None,
                        name=part.name,
                        response={"output": str(part.value)},
                    )
                )
            )
        elif part.type == "thought":
            if encode_thoughts:
                result.append(
                    genai_types.PartDict(text="**Thinking:** " + part.thought)
                )
        else:
            raise NotImplementedError(f"Unsupported content type: {part.type}")
    return result


def _encode_message(
    message: UserMessage | AssistantMessage,
    model_id: GoogleModelId,
    encode_thoughts: bool,
) -> genai_types.ContentDict:
    """Returns a Google `ContentDict` converted from a Mirascope `Message`"""
    if (
        message.role == "assistant"
        and message.provider == "google"
        and message.model_id == model_id
        and message.raw_message
        and not encode_thoughts
    ):
        return cast(genai_types.ContentDict, message.raw_message)
    return genai_types.ContentDict(
        role="model" if message.role == "assistant" else message.role,
        parts=_encode_content(message.content, encode_thoughts),
    )


def _encode_messages(
    messages: Sequence[UserMessage | AssistantMessage],
    model_id: GoogleModelId,
    encode_thoughts: bool,
) -> genai_types.ContentListUnionDict:
    """Returns a `ContentListUnionDict` converted from a sequence of user or assistant `Messages`s"""
    return [_encode_message(message, model_id, encode_thoughts) for message in messages]


@lru_cache(maxsize=128)
def _convert_tool_to_function_declaration(
    tool: ToolSchema,
) -> genai_types.FunctionDeclarationDict:
    """Convert a single Mirascope tool to Google FunctionDeclaration format with caching."""
    schema_dict = tool.parameters.model_dump(by_alias=True, exclude_none=True)
    schema_dict["type"] = "object"

    defs = schema_dict.get("$defs")
    properties = schema_dict.get("properties", {})
    if defs:
        properties = _resolve_refs(properties, defs)

    return genai_types.FunctionDeclarationDict(
        name=tool.name,
        description=tool.description,
        parameters=genai_types.SchemaDict(
            type=genai_types.Type.OBJECT,
            properties=properties,
            required=schema_dict.get("required", []),
        ),
    )


def encode_request(
    *,
    model_id: GoogleModelId,
    messages: Sequence[Message],
    tools: Sequence[ToolSchema] | BaseToolkit | None,
    format: type[FormattableT] | Format[FormattableT] | None,
    params: Params,
) -> tuple[
    Sequence[Message],
    Format[FormattableT] | None,
    genai_types.ContentListUnionDict,
    GoogleKwargs,
]:
    """Prepares a request for the genai `Client.models.generate_content` method."""
    google_config: GoogleKwargs = GoogleKwargs()
    encode_thoughts = False

    with _base_utils.ensure_all_params_accessed(
        params=params, provider="google"
    ) as param_accessor:
        if param_accessor.temperature is not None:
            google_config["temperature"] = param_accessor.temperature
        if param_accessor.max_tokens is not None:
            google_config["max_output_tokens"] = param_accessor.max_tokens
        if param_accessor.top_p is not None:
            google_config["top_p"] = param_accessor.top_p
        if param_accessor.top_k is not None:
            google_config["top_k"] = param_accessor.top_k
        if param_accessor.seed is not None:
            google_config["seed"] = param_accessor.seed
        if param_accessor.stop_sequences is not None:
            google_config["stop_sequences"] = param_accessor.stop_sequences
        if param_accessor.thinking is not None:
            if param_accessor.thinking:
                google_config["thinking_config"] = genai_types.ThinkingConfigDict(
                    thinking_budget=-1,  # automatic budget
                    include_thoughts=True,
                )
            else:
                google_config["thinking_config"] = genai_types.ThinkingConfigDict(
                    include_thoughts=False, thinking_budget=0
                )
        if param_accessor.encode_thoughts_as_text:
            encode_thoughts = True

    tools = tools.tools if isinstance(tools, BaseToolkit) else tools or []
    google_tools: list[genai_types.ToolDict] = []

    format = resolve_format(
        format,
        # Google does not support strict outputs when tools are present
        # (Gemini 2.5 will error, 2.0 and below will ignore tools)
        default_mode="strict" if not tools else "tool",
    )
    if format is not None:
        if format.mode in ("strict", "json") and tools:
            raise FeatureNotSupportedError(
                feature=f"formatting_mode:{format.mode} with tools",
                provider="google",
            )

        if format.mode == "strict":
            google_config["response_mime_type"] = "application/json"
            google_config["response_schema"] = format.schema
        elif format.mode == "tool":
            format_tool_schema = _formatting_utils.create_tool_schema(format)
            format_tool = _convert_tool_to_function_declaration(format_tool_schema)
            google_tools.append(
                genai_types.ToolDict(function_declarations=[format_tool])
            )
            function_calling_config = genai_types.FunctionCallingConfigDict(
                mode=genai_types.FunctionCallingConfigMode.ANY
            )
            if not tools:
                function_calling_config["allowed_function_names"] = [FORMAT_TOOL_NAME]

            google_config["tool_config"] = genai_types.ToolConfigDict(
                function_calling_config=function_calling_config
            )
        elif format.mode == "json":
            google_config["response_mime_type"] = "application/json"

        if format.formatting_instructions:
            messages = _base_utils.add_system_instructions(
                messages, format.formatting_instructions
            )

    if tools:
        function_declarations = [
            _convert_tool_to_function_declaration(tool) for tool in tools
        ]
        google_tools.append(
            genai_types.ToolDict(function_declarations=function_declarations)
        )

    if google_tools:
        google_config["tools"] = cast(genai_types.ToolListUnionDict, google_tools)

    system_message_content, remaining_messages = _base_utils.extract_system_message(
        messages
    )

    if system_message_content:
        google_config["system_instruction"] = system_message_content

    return (
        messages,
        format,
        _encode_messages(remaining_messages, model_id, encode_thoughts),
        google_config,
    )



"""Utilities for the formatting module."""

import inspect
import json

from ..tools import FORMAT_TOOL_NAME, ToolParameterSchema, ToolSchema
from .types import Format, FormattableT, FormattingMode

TOOL_MODE_INSTRUCTIONS = f"""Always respond to the user's query using the {FORMAT_TOOL_NAME} tool for structured output."""


JSON_MODE_INSTRUCTIONS = (
    "Respond only with valid JSON that matches this exact schema:\n{json_schema}"
)


def default_formatting_instructions(
    schema: dict[str, object], mode: FormattingMode
) -> str | None:
    """Generate formatting instructions for the given mode and format info."""

    if mode == "tool":
        return TOOL_MODE_INSTRUCTIONS
    elif mode == "json":
        json_schema = json.dumps(schema, indent=2)
        instructions = JSON_MODE_INSTRUCTIONS.format(json_schema=json_schema)
        return inspect.cleandoc(instructions)


def create_tool_schema(format: Format[FormattableT]) -> ToolSchema:
    """Convert a `Format` to a `ToolSchema` for format parsing.

    Args:
        format: The `Format` instance containing schema and metadata

    Returns:
        `ToolSchema` for the format tool
    """

    schema_dict = format.schema.copy()
    schema_dict["type"] = "object"

    properties = schema_dict.get("properties")
    if not properties or not isinstance(properties, dict):
        properties = {}  # pragma: no cover
    required = list(properties.keys())

    description = (
        f"Use this tool to extract data in {format.name} format for a final response."
    )
    if format.description:
        description += "\n" + format.description

    parameters = ToolParameterSchema(
        properties=properties,
        required=required,
        additionalProperties=False,
    )
    if "$defs" in schema_dict and isinstance(schema_dict["$defs"], dict):
        parameters.defs = schema_dict["$defs"]

    def _unused_format_fn() -> None:
        raise TypeError(
            "Format tool function should not be called."
        )  # pragma: no cover

    tool_schema = ToolSchema.__new__(ToolSchema)
    tool_schema.fn = _unused_format_fn
    tool_schema.name = FORMAT_TOOL_NAME
    tool_schema.description = description
    tool_schema.parameters = parameters
    tool_schema.strict = True

    return tool_schema



"""Tests for the Model class."""

from unittest.mock import MagicMock

from mirascope import llm


def test_client_pulled_from_context_at_call_time() -> None:
    """Test that models get client at call time, not construction time."""
    model = llm.use_model(provider="openai:completions", model_id="gpt-4o")

    custom_client = llm.client("openai:completions", api_key="test-key")

    mock_call = MagicMock(return_value=MagicMock())
    custom_client.call = mock_call

    with custom_client:
        model.call(messages=[llm.messages.user("Hello")])

    mock_call.assert_called_once()


def test_use_model_without_context() -> None:
    """Test that use_model creates a new Model when no context is set."""
    model = llm.use_model(provider="openai:completions", model_id="gpt-4o")

    assert model.provider == "openai:completions"
    assert model.model_id == "gpt-4o"


def test_use_model_with_context() -> None:
    """Test that use_model returns the context model when one is set."""
    with llm.model(provider="anthropic", model_id="claude-sonnet-4-0"):
        model = llm.use_model(provider="openai:completions", model_id="gpt-4o")

        assert model.provider == "anthropic"
        assert model.model_id == "claude-sonnet-4-0"


def test_direct_model_instantiation_ignores_context() -> None:
    """Test that direct Model instantiation ignores context (hardcoding behavior)."""
    with llm.model(provider="anthropic", model_id="claude-sonnet-4-0"):
        model = llm.Model(provider="openai:completions", model_id="gpt-4o")

        assert model.provider == "openai:completions"
        assert model.model_id == "gpt-4o"



"""Tests for llm.clients.AnthropicClient."""

from mirascope import llm


def test_context_manager() -> None:
    """Test nested context manager behavior and get_client() integration."""

    global_client = llm.get_client("anthropic")

    client1 = llm.client("anthropic", api_key="key1")
    client2 = llm.client("anthropic", api_key="key2")

    assert llm.get_client("anthropic") is global_client

    with client1 as ctx1:
        assert ctx1 is client1
        assert llm.get_client("anthropic") is client1

        with client2 as ctx2:
            assert ctx2 is client2
            assert llm.get_client("anthropic") is client2

        assert llm.get_client("anthropic") is client1

    assert llm.get_client("anthropic") is global_client


def test_client_caching() -> None:
    """Test that client() returns cached instances for identical parameters."""
    client1 = llm.client(
        "anthropic", api_key="test-key", base_url="https://api.example.com"
    )
    client2 = llm.client(
        "anthropic", api_key="test-key", base_url="https://api.example.com"
    )
    assert client1 is client2

    client3 = llm.client(
        "anthropic", api_key="different-key", base_url="https://api.example.com"
    )
    assert client1 is not client3

    client4 = llm.client(
        "anthropic", api_key="test-key", base_url="https://different.example.com"
    )
    assert client1 is not client4

    client5 = llm.client("anthropic", api_key=None, base_url=None)
    client6 = llm.get_client("anthropic")
    assert client5 is client6



"""Tests for the @tool decorator."""

from typing import Annotated

import pytest
from pydantic import Field

from mirascope import llm


class TestTool:
    """Tests for regular tool decorator (no context dependency)."""

    def test_sync_tool_basic(self) -> None:
        @llm.tool
        def add_numbers(a: int, b: int) -> int:
            """Add two numbers together."""
            return a + b

        assert isinstance(add_numbers, llm.Tool)
        assert add_numbers.name == "add_numbers"
        assert add_numbers.description == "Add two numbers together."
        assert add_numbers.strict is False

        params = add_numbers.parameters
        assert "a" in params.properties
        assert "b" in params.properties
        assert params.required == ["a", "b"]

        assert add_numbers(2, 3) == 5

        tool_call = llm.ToolCall(
            id="call_123", name="add_numbers", args='{"a": 5, "b": 7}'
        )
        output = add_numbers.execute(tool_call)
        expected = llm.ToolOutput(id="call_123", value=12, name="add_numbers")
        assert output == expected

    def test_sync_tool_with_strict(self) -> None:
        @llm.tool(strict=True)
        def multiply(x: int, y: int = 2) -> int:
            """Multiply two numbers."""
            return x * y

        assert isinstance(multiply, llm.Tool)
        assert multiply.strict is True
        assert multiply.name == "multiply"

    def test_sync_tool_with_annotated_descriptions(self) -> None:
        @llm.tool
        def get_weather(
            location: Annotated[str, Field(description="The city and state")],
            unit: Annotated[str, Field(description="Temperature unit")] = "fahrenheit",
        ) -> str:
            """Get weather for a location."""
            return f"Weather in {location} is sunny, 75°{unit[0].upper()}"

        assert isinstance(get_weather, llm.Tool)

        params = get_weather.parameters
        assert params.properties["location"]["description"] == "The city and state"
        assert params.properties["unit"]["description"] == "Temperature unit"

        result = get_weather("San Francisco, CA", "celsius")
        assert "San Francisco" in result and "75°C" in result

    @pytest.mark.asyncio
    async def test_async_tool_basic(self) -> None:
        @llm.tool
        async def async_add(a: int, b: int) -> int:
            """Add two numbers asynchronously."""
            return a + b

        assert isinstance(async_add, llm.AsyncTool)
        assert async_add.name == "async_add"
        assert async_add.description == "Add two numbers asynchronously."

        assert await async_add(3, 4) == 7

        tool_call = llm.ToolCall(
            id="call_456", name="async_add", args='{"a": 10, "b": 15}'
        )
        output = await async_add.execute(tool_call)
        expected = llm.ToolOutput(id="call_456", value=25, name="async_add")
        assert output == expected

    def test_decorator_syntax(self) -> None:
        @llm.tool
        def without_parens() -> str:
            """A simple tool."""
            return "simple"

        @llm.tool()
        def with_parens() -> str:
            """Another simple tool."""
            return "another"

        assert isinstance(without_parens, llm.Tool)
        assert without_parens.name == "without_parens"
        assert without_parens() == "simple"

        assert isinstance(with_parens, llm.Tool)
        assert with_parens.name == "with_parens"
        assert with_parens() == "another"

    def test_decorator_modes_equivalence(self) -> None:
        """Test that @tool and tool()(fn) produce equivalent results."""

        def base_func(x: int, y: str = "default") -> str:
            """A function to be decorated."""
            return f"{y}: {x}"

        decorated = llm.tool(base_func)
        via_factory = llm.tool()(base_func)

        assert decorated == via_factory

    def test_decorator_modes_equivalence_async(self) -> None:
        """Test that @tool and tool()(fn) produce equivalent results."""

        async def base_func(x: int, y: str = "default") -> str:
            """A function to be decorated."""
            return f"{y}: {x}"

        decorated = llm.tool(base_func)
        via_factory = llm.tool()(base_func)

        assert decorated == via_factory


class TestContextTool:
    """Tests for context tool decorator (with context dependency)."""

    def test_sync_context_tool_basic(self) -> None:
        @llm.tool
        def add_with_constant(ctx: llm.Context[int], a: int) -> int:
            """Add number to constant from context."""
            return a + ctx.deps

        assert isinstance(add_with_constant, llm.ContextTool)
        assert add_with_constant.name == "add_with_constant"
        assert add_with_constant.description == "Add number to constant from context."
        assert add_with_constant.strict is False

        params = add_with_constant.parameters
        assert "a" in params.properties
        assert "ctx" not in params.properties
        assert params.required == ["a"]

        context = llm.Context[int](deps=10)
        assert add_with_constant(context, 5) == 15

        tool_call = llm.ToolCall(
            id="call_123", name="add_with_constant", args='{"a": 7}'
        )
        output = add_with_constant.execute(context, tool_call)
        expected = llm.ToolOutput(id="call_123", value=17, name="add_with_constant")
        assert output == expected

    def test_sync_context_tool_with_strict(self) -> None:
        @llm.tool(strict=True)
        def multiply_with_base(ctx: llm.Context[int], x: int, y: int = 2) -> int:
            """Multiply two numbers and add base from context."""
            return (x * y) + ctx.deps

        assert isinstance(multiply_with_base, llm.ContextTool)
        assert multiply_with_base.strict is True
        assert multiply_with_base.name == "multiply_with_base"

    def test_sync_context_tool_with_annotated_descriptions(self) -> None:
        @llm.tool
        def get_weather_with_api(
            ctx: llm.Context[str],
            location: Annotated[str, Field(description="The city and state")],
            unit: Annotated[str, Field(description="Temperature unit")] = "fahrenheit",
        ) -> str:
            """Get weather for a location using API key from context."""
            return f"Weather in {location} is sunny, 75°{unit[0].upper()} (API: {ctx.deps})"

        assert isinstance(get_weather_with_api, llm.ContextTool)

        params = get_weather_with_api.parameters
        assert params.properties["location"]["description"] == "The city and state"
        assert params.properties["unit"]["description"] == "Temperature unit"

        context = llm.Context[str](deps="secret-api-key")
        result = get_weather_with_api(context, "San Francisco, CA", "celsius")
        assert (
            "San Francisco" in result
            and "75°C" in result
            and "secret-api-key" in result
        )

    @pytest.mark.asyncio
    async def test_async_context_tool_basic(self) -> None:
        @llm.tool
        async def async_add_with_base(ctx: llm.Context[int], a: int) -> int:
            """Add number to base from context asynchronously."""
            return a + ctx.deps

        assert isinstance(async_add_with_base, llm.AsyncContextTool)
        assert async_add_with_base.name == "async_add_with_base"
        assert (
            async_add_with_base.description
            == "Add number to base from context asynchronously."
        )

        context = llm.Context[int](deps=15)
        assert await async_add_with_base(context, 10) == 25

        tool_call = llm.ToolCall(
            id="call_456", name="async_add_with_base", args='{"a": 20}'
        )
        output = await async_add_with_base.execute(context, tool_call)
        expected = llm.ToolOutput(id="call_456", value=35, name="async_add_with_base")
        assert output == expected

    def test_decorator_syntax(self) -> None:
        @llm.tool
        def without_parens(ctx: llm.Context[int]) -> str:
            """A simple context tool."""
            return f"simple-{ctx.deps}"

        @llm.tool()
        def with_parens(ctx: llm.Context[int]) -> str:
            """Another simple context tool."""
            return f"another-{ctx.deps}"

        assert isinstance(without_parens, llm.ContextTool)
        assert without_parens.name == "without_parens"

        context = llm.Context[int](deps=42)
        assert without_parens(context) == "simple-42"

        assert isinstance(with_parens, llm.ContextTool)
        assert with_parens.name == "with_parens"
        assert with_parens(context) == "another-42"

    def test_decorator_modes_equivalence(self) -> None:
        """Test that @context_tool and context_tool()(fn) produce equivalent results."""

        def base_func(ctx: llm.Context[int], x: int, y: str = "default") -> str:
            """A function to be decorated."""
            return f"{y}: {x + ctx.deps}"

        decorated = llm.tool(base_func)
        via_factory = llm.tool()(base_func)

        assert decorated == via_factory

    def test_decorator_modes_equivalence_async(self) -> None:
        """Test that @context_tool and context_tool()(fn) produce equivalent results for async."""

        async def base_func(ctx: llm.Context[int], x: int, y: str = "default") -> str:
            """A function to be decorated."""
            return f"{y}: {x + ctx.deps}"

        decorated = llm.tool(base_func)
        via_factory = llm.tool()(base_func)

        assert decorated == via_factory

    def test_parameter_filtering(self) -> None:
        """Test that context parameter is filtered out of tool schema."""

        @llm.tool
        def test_func(
            ctx: llm.Context[str], param1: int, param2: str = "default"
        ) -> str:
            """Test function with context and regular parameters."""
            return f"{param2}-{param1}-{ctx.deps}"

        params = test_func.parameters

        assert set(params.properties.keys()) == {"param1", "param2"}
        assert "ctx" not in params.properties
        assert params.required == ["param1"]

    def test_context_parameter_name_must_be_ctx(self) -> None:
        """Test that context tools work require the first parameter be named `ctx`."""

        @llm.tool
        def context_weird_name(special_context: llm.Context[str], value: int) -> str:
            return f"{special_context.deps}-{value}"

        assert isinstance(context_weird_name, llm.Tool)
        assert "special_context" in context_weird_name.parameters.properties
        assert "value" in context_weird_name.parameters.properties

    def test_async_context_parameter_name_must_be_ctx(self) -> None:
        """Test that async context tools work require the first parameter be named `ctx`."""

        @llm.tool
        async def context_weird_name(
            special_context: llm.Context[str], value: int
        ) -> str:
            return f"{special_context.deps}-{value}"

        assert isinstance(context_weird_name, llm.AsyncTool)
        assert "special_context" in context_weird_name.parameters.properties
        assert "value" in context_weird_name.parameters.properties

    def test_untyped_parameter(self) -> None:
        """Test that untyped ctx parameters don't become context tools."""

        @llm.tool
        def non_context_tool(ctx) -> str:  # noqa: ANN001
            return f"{ctx}"

        assert isinstance(non_context_tool, llm.Tool)
        assert "ctx" in non_context_tool.parameters.properties

    def test_async_untyped_parameter(self) -> None:
        """Test that untyped ctx parameters don't become context tools."""

        @llm.tool
        async def non_context_tool(ctx) -> str:  # noqa: ANN001
            return f"{ctx}"

        assert isinstance(non_context_tool, llm.AsyncTool)
        assert "ctx" in non_context_tool.parameters.properties

    def test_non_context_typed_parameter(self) -> None:
        """Test that non-Context typed parameters are not treated as context tools."""

        @llm.tool
        def non_context_tool(ctx: int) -> str:
            return f"{ctx}"

        assert isinstance(non_context_tool, llm.Tool)
        assert "ctx" in non_context_tool.parameters.properties

    def test_async_non_context_typed_parameter(self) -> None:
        """Test that non-Context typed async parameters are not treated as context tools."""

        @llm.tool
        async def non_context_tool(ctx: int) -> str:
            return f"{ctx}"

        assert isinstance(non_context_tool, llm.AsyncTool)
        assert "ctx" in non_context_tool.parameters.properties

    def test_method_with_context_after_self(self) -> None:
        """Test that methods with self as first arg and Context as second arg are context tools."""

        class TestClass:
            @llm.tool
            def method_with_context(self, ctx: llm.Context[str], value: int) -> str:
                return f"{ctx.deps}-{value}"

        assert isinstance(TestClass.method_with_context, llm.ContextTool)
        assert "ctx" not in TestClass.method_with_context.parameters.properties
        assert "value" in TestClass.method_with_context.parameters.properties

    def test_async_method_with_context_after_self(self) -> None:
        """Test that async methods with self as first arg and Context as second arg are context tools."""

        class TestClass:
            @llm.tool
            async def method_with_context(
                self, ctx: llm.Context[str], value: int
            ) -> str:
                return f"{ctx.deps}-{value}"

        assert isinstance(TestClass.method_with_context, llm.AsyncContextTool)
        assert "ctx" not in TestClass.method_with_context.parameters.properties
        assert "value" in TestClass.method_with_context.parameters.properties

    def test_context_not_first_parameter(self) -> None:
        """Test that Context as second parameter (after non-self) is not treated as context tool."""

        @llm.tool
        def second_arg_context(regular_param: int, ctx: llm.Context[str]) -> str:
            return f"{regular_param}-{ctx.deps}"

        assert isinstance(second_arg_context, llm.Tool)
        assert "regular_param" in second_arg_context.parameters.properties
        assert "ctx" in second_arg_context.parameters.properties

    def test_async_context_not_first_parameter(self) -> None:
        """Test that Context as second parameter (after non-self) is not treated as async context tool."""

        @llm.tool
        async def second_arg_context(regular_param: int, ctx: llm.Context[str]) -> str:
            return f"{regular_param}-{ctx.deps}"

        assert isinstance(second_arg_context, llm.AsyncTool)
        assert "regular_param" in second_arg_context.parameters.properties
        assert "ctx" in second_arg_context.parameters.properties

    def test_context_subclass_detection(self) -> None:
        """Test that Context subclasses are properly detected as context tools."""

        class CustomContext(llm.Context[str]): ...

        @llm.tool
        def with_custom_context(ctx: CustomContext, value: int) -> str:
            return str(value)

        assert isinstance(with_custom_context, llm.ContextTool)
        assert "ctx" not in with_custom_context.parameters.properties
        assert "value" in with_custom_context.parameters.properties

    def test_async_context_subclass_detection(self) -> None:
        """Test that Context subclasses are properly detected as async context tools."""

        class CustomContext(llm.Context[str]): ...

        @llm.tool
        async def with_custom_context(ctx: CustomContext, value: int) -> str:
            return str(value)

        assert isinstance(with_custom_context, llm.AsyncContextTool)
        assert "ctx" not in with_custom_context.parameters.properties
        assert "value" in with_custom_context.parameters.properties



"""Anthropic response decoding."""

import json
from typing import Any, TypeAlias, cast

from anthropic import types as anthropic_types
from anthropic.lib.streaming import AsyncMessageStreamManager, MessageStreamManager

from ....content import (
    AssistantContentPart,
    Text,
    TextChunk,
    TextEndChunk,
    TextStartChunk,
    Thought,
    ThoughtChunk,
    ThoughtEndChunk,
    ThoughtStartChunk,
    ToolCall,
    ToolCallChunk,
    ToolCallEndChunk,
    ToolCallStartChunk,
)
from ....messages import AssistantMessage
from ....responses import (
    AsyncChunkIterator,
    ChunkIterator,
    FinishReason,
    FinishReasonChunk,
    RawMessageChunk,
    RawStreamEventChunk,
)
from ..model_ids import AnthropicModelId

ANTHROPIC_FINISH_REASON_MAP = {
    "max_tokens": FinishReason.MAX_TOKENS,
    "refusal": FinishReason.REFUSAL,
}


def _decode_assistant_content(
    content: anthropic_types.ContentBlock,
) -> AssistantContentPart:
    """Convert Anthropic content block to mirascope AssistantContentPart."""
    if content.type == "text":
        return Text(text=content.text)
    elif content.type == "tool_use":
        return ToolCall(
            id=content.id,
            name=content.name,
            args=json.dumps(content.input),
        )
    elif content.type == "thinking":
        return Thought(thought=content.thinking)
    else:
        raise NotImplementedError(
            f"Support for content type `{content.type}` is not yet implemented."
        )


def decode_response(
    response: anthropic_types.Message,
    model_id: AnthropicModelId,
) -> tuple[AssistantMessage, FinishReason | None]:
    """Convert Anthropic message to mirascope AssistantMessage."""
    assistant_message = AssistantMessage(
        content=[_decode_assistant_content(part) for part in response.content],
        provider="anthropic",
        model_id=model_id,
        raw_message={
            "role": response.role,
            "content": [part.model_dump() for part in response.content],
        },
    )
    finish_reason = (
        ANTHROPIC_FINISH_REASON_MAP.get(response.stop_reason)
        if response.stop_reason
        else None
    )
    return assistant_message, finish_reason


ContentBlock: TypeAlias = (
    anthropic_types.TextBlockParam
    | anthropic_types.ThinkingBlockParam
    | anthropic_types.ToolUseBlockParam
    | anthropic_types.ThinkingBlockParam
    | anthropic_types.RedactedThinkingBlockParam
)


class _AnthropicChunkProcessor:
    """Processes Anthropic stream events and maintains state across events."""

    def __init__(self) -> None:
        self.current_block_param: ContentBlock | None = None
        self.accumulated_tool_json: str = ""
        self.accumulated_blocks: list[ContentBlock] = []

    def process_event(
        self, event: anthropic_types.RawMessageStreamEvent
    ) -> ChunkIterator:
        """Process a single Anthropic event and yield the appropriate content chunks."""
        yield RawStreamEventChunk(raw_stream_event=event)

        if event.type == "content_block_start":
            content_block = event.content_block

            if content_block.type == "text":
                self.current_block_param = {
                    "type": "text",
                    "text": content_block.text,
                }
                yield TextStartChunk()
            elif content_block.type == "tool_use":
                self.current_block_param = {
                    "type": "tool_use",
                    "id": content_block.id,
                    "name": content_block.name,
                    "input": {},
                }
                self.accumulated_tool_json = ""
                yield ToolCallStartChunk(
                    id=content_block.id,
                    name=content_block.name,
                )
            elif content_block.type == "thinking":
                self.current_block_param = {
                    "type": "thinking",
                    "thinking": "",
                    "signature": "",
                }
                yield ThoughtStartChunk()
            elif content_block.type == "redacted_thinking":  # pragma: no cover
                self.current_block_param = {
                    "type": "redacted_thinking",
                    "data": content_block.data,
                }
            else:
                raise NotImplementedError

        elif event.type == "content_block_delta":
            if self.current_block_param is None:  # pragma: no cover
                raise RuntimeError("Received delta without a current block")

            delta = event.delta
            if delta.type == "text_delta":
                if self.current_block_param["type"] != "text":  # pragma: no cover
                    raise RuntimeError(
                        f"Received text_delta for {self.current_block_param['type']} block"
                    )
                self.current_block_param["text"] += delta.text
                yield TextChunk(delta=delta.text)
            elif delta.type == "input_json_delta":
                if self.current_block_param["type"] != "tool_use":  # pragma: no cover
                    raise RuntimeError(
                        f"Received input_json_delta for {self.current_block_param['type']} block"
                    )
                self.accumulated_tool_json += delta.partial_json
                yield ToolCallChunk(delta=delta.partial_json)
            elif delta.type == "thinking_delta":
                if self.current_block_param["type"] != "thinking":  # pragma: no cover
                    raise RuntimeError(
                        f"Received thinking_delta for {self.current_block_param['type']} block"
                    )
                self.current_block_param["thinking"] += delta.thinking
                yield ThoughtChunk(delta=delta.thinking)
            elif delta.type == "signature_delta":
                if self.current_block_param["type"] != "thinking":  # pragma: no cover
                    raise RuntimeError(
                        f"Received signature_delta for {self.current_block_param['type']} block"
                    )
                self.current_block_param["signature"] += delta.signature
            else:
                raise RuntimeError(
                    f"Received unsupported delta type: {delta.type}"
                )  # pragma: no cover

        elif event.type == "content_block_stop":
            if self.current_block_param is None:  # pragma: no cover
                raise RuntimeError("Received stop without a current block")

            block_type = self.current_block_param["type"]

            if block_type == "text":
                yield TextEndChunk()
            elif block_type == "tool_use":
                if self.current_block_param["type"] != "tool_use":  # pragma: no cover
                    raise RuntimeError(
                        f"Block type mismatch: stored {self.current_block_param['type']}, expected tool_use"
                    )
                self.current_block_param["input"] = (
                    json.loads(self.accumulated_tool_json)
                    if self.accumulated_tool_json
                    else {}
                )
                yield ToolCallEndChunk()
            elif block_type == "thinking":
                yield ThoughtEndChunk()
            else:
                raise NotImplementedError

            self.accumulated_blocks.append(self.current_block_param)
            self.current_block_param = None

        elif event.type == "message_delta":
            if event.delta.stop_reason:
                finish_reason = ANTHROPIC_FINISH_REASON_MAP.get(event.delta.stop_reason)
                if finish_reason is not None:
                    yield FinishReasonChunk(finish_reason=finish_reason)

    def raw_message_chunk(self) -> RawMessageChunk:
        return RawMessageChunk(
            raw_message=cast(
                dict[str, Any],
                {
                    "role": "assistant",
                    "content": self.accumulated_blocks,
                },
            )
        )


def decode_stream(
    anthropic_stream_manager: MessageStreamManager,
) -> ChunkIterator:
    """Returns a ChunkIterator converted from an Anthropic MessageStreamManager"""
    processor = _AnthropicChunkProcessor()
    with anthropic_stream_manager as stream:
        for event in stream._raw_stream:
            yield from processor.process_event(event)
    yield processor.raw_message_chunk()


async def decode_async_stream(
    anthropic_stream_manager: AsyncMessageStreamManager,
) -> AsyncChunkIterator:
    """Returns an AsyncChunkIterator converted from an Anthropic MessageStreamManager"""
    processor = _AnthropicChunkProcessor()
    async with anthropic_stream_manager as stream:
        async for event in stream._raw_stream:
            for item in processor.process_event(event):
                yield item
    yield processor.raw_message_chunk()



"""Shared utils for all OpenAI clients."""

MODELS_WITHOUT_JSON_SCHEMA_SUPPORT = {
    "chatgpt-4o-latest",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-0125",
    "gpt-3.5-turbo-1106",
    "gpt-3.5-turbo-16k",
    "gpt-4",
    "gpt-4-0125-preview",
    "gpt-4-0613",
    "gpt-4-1106-preview",
    "gpt-4-turbo",
    "gpt-4-turbo-2024-04-09",
    "gpt-4-turbo-preview",
    "gpt-4o-2024-05-13",
    "gpt-4o-audio-preview",
    "gpt-4o-audio-preview-2024-10-01",
    "gpt-4o-audio-preview-2024-12-17",
    "gpt-4o-audio-preview-2025-06-03",
    "gpt-4o-mini-audio-preview",
    "gpt-4o-mini-audio-preview-2024-12-17",
    "gpt-5-chat-latest",
    "o1-mini",
    "o1-mini-2024-09-12",
}

MODELS_WITHOUT_JSON_OBJECT_SUPPORT = {
    "gpt-4",
    "gpt-4-0613",
    "gpt-4o-audio-preview",
    "gpt-4o-audio-preview-2024-10-01",
    "gpt-4o-audio-preview-2024-12-17",
    "gpt-4o-audio-preview-2025-06-03",
    "gpt-4o-mini-audio-preview",
    "gpt-4o-mini-audio-preview-2024-12-17",
    "gpt-4o-mini-search-preview",
    "gpt-4o-mini-search-preview-2025-03-11",
    "gpt-4o-search-preview",
    "gpt-4o-search-preview-2025-03-11",
    "o1-mini",
    "o1-mini-2024-09-12",
}


def _ensure_additional_properties_false(obj: object) -> None:
    """Recursively adds additionalProperties = False to a schema, required by OpenAI API."""
    if isinstance(obj, dict):
        if obj.get("type") == "object" and "additionalProperties" not in obj:
            obj["additionalProperties"] = False
        for value in obj.values():
            _ensure_additional_properties_false(value)
    elif isinstance(obj, list):
        for item in obj:
            _ensure_additional_properties_false(item)



"""Types for prompt functions."""

from typing import Protocol, TypeVar

from ..context import Context, DepsT
from ..messages import Message, UserContent
from ..types import P

PromptT = TypeVar(
    "PromptT", bound="Prompt | AsyncPrompt | ContextPrompt | AsyncContextPrompt"
)
"""Type variable for prompt types.

This type var represents a resolved prompt, i.e. one that returns a list of messages.
"""


class Prompt(Protocol[P]):
    """Protocol for a `Prompt`, which returns `list[Message]`."""

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> list[Message]: ...


class Promptable(Protocol[P]):
    """Protocol for a `Promptable` that returns `UserContent` or `list[Message]`.

    May be be converted by the `prompt` decorator into a `Prompt`.
    """

    def __call__(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> UserContent | list[Message]: ...


class AsyncPrompt(Protocol[P]):
    """Protocol for an `AsyncPrompt`, which returns `list[Message]`."""

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> list[Message]: ...


class AsyncPromptable(Protocol[P]):
    """Protocol for an `AsyncPromptable` that returns `UserContent` or `list[Message]`.

    May be converted by the `prompt` decorator into an `AsyncPrompt`.
    """

    async def __call__(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> UserContent | list[Message]: ...


class ContextPrompt(Protocol[P, DepsT]):
    """Protocol for a `ContextPrompt`, which returns `list[Message]`."""

    def __call__(
        self,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> list[Message]: ...


class ContextPromptable(Protocol[P, DepsT]):
    """Protocol for a `ContextPromptable` that returns `UserContent` or `list[Message]`.

    May be converted by the `prompt` decorator into a `ContextPrompt`.
    """

    def __call__(
        self,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> UserContent | list[Message]: ...


class AsyncContextPrompt(Protocol[P, DepsT]):
    """Protocol for an `AsyncContextPrompt`, which returns `list[Message]`."""

    async def __call__(
        self,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> list[Message]: ...


class AsyncContextPromptable(Protocol[P, DepsT]):
    """Protocol for an `AsyncContextPromptable` that returns `UserContent` or `list[Message]`.

    May be converted by the `prompt` decorator into an `AsyncContextPrompt`.
    """

    async def __call__(
        self,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> UserContent | list[Message]: ...



"""Configuration for all Mirascope tests."""

from __future__ import annotations

import os

import pytest
from dotenv import load_dotenv


@pytest.fixture(scope="session", autouse=True)
def load_api_keys() -> None:
    """Load environment variables from .env file.

    This is necessary for e2e tests, but also may be necessary for any tests that
    instantiate a client, as clients may test API key presence at `__init__` time.
    """
    load_dotenv()
    # Set dummy keys if not present so that tests pass in CI.
    os.environ.setdefault("ANTHROPIC_API_KEY", "dummy-anthropic-key")
    os.environ.setdefault("GOOGLE_API_KEY", "dummy-google-key")
    os.environ.setdefault("OPENAI_API_KEY", "dummy-openai-key")



from mirascope import llm


@llm.call(provider="openai", model_id="gpt-5")
def recommend_book(genre: str):
    return [
        # [!code highlight]
        llm.messages.system("Always recommend kid-friendly books."),
        llm.messages.user(f"Please recommend a book in {genre}."),
    ]


def main():
    response: llm.Response = recommend_book("fantasy")
    print(response.pretty())


main()



from inline_snapshot import snapshot

test_snapshot = snapshot(
    {
        "exception": {
            "type": "FeatureNotSupportedError",
            "args": "(\"Model 'gpt-4o' does not support audio inputs.\",)",
            "feature": "Audio inputs",
            "model_id": "None",
            "provider": "openai:completions",
        }
    }
)



"""Base interface for all LLM responses."""

from abc import ABC
from collections.abc import Sequence
from types import NoneType
from typing import TYPE_CHECKING, Any, Generic, Literal, overload

from ..content import AssistantContentPart, Text, Thought, ToolCall
from ..formatting import Format, FormattableT, Partial
from ..messages import Message
from ..tools import ToolkitT
from . import _utils
from .finish_reason import FinishReason

if TYPE_CHECKING:
    from ..clients import ModelId, Params, Provider
    from ..models import Model


class RootResponse(Generic[ToolkitT, FormattableT], ABC):
    """Base class for LLM responses."""

    raw: Any
    """The raw response from the LLM."""

    provider: "Provider"
    """The provider that generated this response."""

    model_id: "ModelId"
    """The model id that generated this response."""

    params: "Params"
    """The params that were used to generate this response (or None)."""

    toolkit: ToolkitT
    """The toolkit containing the tools used when generating this response."""

    messages: list[Message]
    """The message history, including the most recent assistant message."""

    content: Sequence[AssistantContentPart]
    """The content generated by the LLM."""

    texts: Sequence[Text]
    """The text content in the generated response, if any."""

    tool_calls: Sequence[ToolCall]
    """The tools the LLM wants called on its behalf, if any."""

    thoughts: Sequence[Thought]
    """The readable thoughts from the model's thinking process, if any.

    The thoughts may be direct output from the model thinking process, or may be a
    generated summary. (This depends on the provider; newer models tend to summarize.)
    """
    finish_reason: FinishReason | None
    """The reason why the LLM finished generating a response, if set.
    
    `finish_reason` is only set if the response did not finish generating normally,
    e.g. `FinishReason.MAX_TOKENS` if the model ran out of tokens before completing.
    When the response generates normally, `response.finish_reason` will be `None`.
    """

    format: Format[FormattableT] | None
    """The `Format` describing the structured response format, if available."""

    @overload
    def parse(self: "RootResponse[ToolkitT, None]", partial: Literal[True]) -> None:
        """Format the response into a `Partial[BaseModel]` (with optional fields).

        This is useful for when the stream is only partially consumed, in which case the
        structured output may only be partially available.
        """
        ...

    @overload
    def parse(
        self: "RootResponse[ToolkitT, FormattableT]", partial: Literal[True]
    ) -> Partial[FormattableT]:
        """Format the response into a `Partial[BaseModel]` (with optional fields).

        This is useful for when the stream is only partially consumed, in which case the
        structured output may only be partially available.
        """
        ...

    @overload
    def parse(
        self: "RootResponse[ToolkitT, None]", partial: Literal[False] = False
    ) -> None:
        """Overload when the format type is `None`."""
        ...

    @overload
    def parse(
        self: "RootResponse[ToolkitT, FormattableT]", partial: Literal[False] = False
    ) -> FormattableT:
        """Overload when the format type is not `None`."""
        ...

    def parse(
        self, partial: bool = False
    ) -> FormattableT | Partial[FormattableT] | None:
        """Format the response according to the response format parser.

        Returns:
            The formatted response object of type FormatT.

        Raises:
            json.JSONDecodeError: If the response's textual content can't be parsed as
                JSON.
            pydantic.ValidationError: If the response's content fails validation for the
                format type.
        """
        if self.format is None:
            return None

        formattable = self.format.formattable
        if formattable is None or formattable is NoneType:
            return None  # pragma: no cover

        if partial:
            raise NotImplementedError

        text = "".join(text.text for text in self.texts)
        json_text = _utils.extract_serialized_json(text)

        return formattable.model_validate_json(json_text)

    def pretty(self) -> str:
        """Return a string representation of all response content.

        The response content will be represented in a way that emphasies clarity and
        readability, but may not include all metadata (like thinking signatures or tool
        call ids), and thus cannot be used to reconstruct the response. For example:

        **Thinking:**
          The user is asking a math problem. I should use the calculator tool.

        **Tool Call (calculator)** {'operation': 'mult', 'a': 1337, 'b': 4242}

        I am going to use the calculator and answer your question for you!
        """
        if not self.content:
            return "**[No Content]**"

        pretty_parts: list[str] = []
        for part in self.content:
            if isinstance(part, Text):
                pretty_parts.append(part.text)
            elif isinstance(part, ToolCall):
                pretty_parts.append(f"**ToolCall ({part.name}):** {part.args}")
            elif isinstance(part, Thought):
                indented_thinking = "\n".join(
                    f"  {line}" for line in part.thought.split("\n")
                )
                pretty_parts.append(f"**Thinking:**\n{indented_thinking}")
            else:
                pretty_parts.append(
                    f"[{type(part).__name__}: {str(part)}]"
                )  # pragma: no cover

        return "\n\n".join(pretty_parts)

    @property
    def model(self) -> "Model":
        """A `Model` with parameters matching this response."""
        from ..models import Model, get_model_from_context

        if context_model := get_model_from_context():
            return context_model

        return Model(
            provider=self.provider,
            model_id=self.model_id,
            **self.params,
        )
